@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004 Choe Hwanjin
+ * Copyright (C) 2003,2004,2007 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -41,13 +41,17 @@ nabi_candidate_on_row_activated(GtkWidget *widget,
 				GtkTreeViewColumn *column,
 				NabiCandidate *candidate)
 {
+    const Hanja* hanja;
     if (path != NULL) {
 	int *indices;
 	indices = gtk_tree_path_get_indices(path);
 	candidate->current = candidate->first + indices[0];
     }
-    if (candidate->commit != NULL && candidate->commit_data != NULL)
-	candidate->commit(candidate, candidate->commit_data);
+
+    hanja = nabi_candidate_get_current(candidate);
+    if (hanja != NULL &&
+	candidate->commit != NULL && candidate->commit_data != NULL)
+	candidate->commit(candidate, hanja, candidate->commit_data);
 }
 
 static void
@@ -85,7 +89,7 @@ nabi_candidate_on_key_press(GtkWidget *widget,
 			    GdkEventKey *event,
 			    NabiCandidate *candidate)
 {
-    wchar_t ch = 0;
+    const Hanja* hanja = NULL;
 
     if (candidate == NULL)
 	return FALSE;
@@ -120,10 +124,10 @@ nabi_candidate_on_key_press(GtkWidget *widget,
 	break;
     case GDK_Return:
     case GDK_KP_Enter:
-	ch = nabi_candidate_get_current(candidate);
+	hanja = nabi_candidate_get_current(candidate);
 	break;
     case GDK_0:
-	ch = nabi_candidate_get_nth(candidate, 9);
+	hanja = nabi_candidate_get_nth(candidate, 9);
 	break;
     case GDK_1:
     case GDK_2:
@@ -134,10 +138,10 @@ nabi_candidate_on_key_press(GtkWidget *widget,
     case GDK_7:
     case GDK_8:
     case GDK_9:
-	ch = nabi_candidate_get_nth(candidate, event->keyval - GDK_1);
+	hanja = nabi_candidate_get_nth(candidate, event->keyval - GDK_1);
 	break;
     case GDK_KP_0:
-	ch = nabi_candidate_get_nth(candidate, 9);
+	hanja = nabi_candidate_get_nth(candidate, 9);
 	break;
     case GDK_KP_1:
     case GDK_KP_2:
@@ -148,45 +152,45 @@ nabi_candidate_on_key_press(GtkWidget *widget,
     case GDK_KP_7:
     case GDK_KP_8:
     case GDK_KP_9:
-	ch = nabi_candidate_get_nth(candidate, event->keyval - GDK_KP_1);
+	hanja = nabi_candidate_get_nth(candidate, event->keyval - GDK_KP_1);
 	break;
     case GDK_KP_End:
-	ch = nabi_candidate_get_nth(candidate, 0);
+	hanja = nabi_candidate_get_nth(candidate, 0);
 	break;
     case GDK_KP_Down:
-	ch = nabi_candidate_get_nth(candidate, 1);
+	hanja = nabi_candidate_get_nth(candidate, 1);
 	break;
     case GDK_KP_Next:
-	ch = nabi_candidate_get_nth(candidate, 2);
+	hanja = nabi_candidate_get_nth(candidate, 2);
 	break;
     case GDK_KP_Left:
-	ch = nabi_candidate_get_nth(candidate, 3);
+	hanja = nabi_candidate_get_nth(candidate, 3);
 	break;
     case GDK_KP_Begin:
-	ch = nabi_candidate_get_nth(candidate, 4);
+	hanja = nabi_candidate_get_nth(candidate, 4);
 	break;
     case GDK_KP_Right:
-	ch = nabi_candidate_get_nth(candidate, 5);
+	hanja = nabi_candidate_get_nth(candidate, 5);
 	break;
     case GDK_KP_Home:
-	ch = nabi_candidate_get_nth(candidate, 6);
+	hanja = nabi_candidate_get_nth(candidate, 6);
 	break;
     case GDK_KP_Up:
-	ch = nabi_candidate_get_nth(candidate, 7);
+	hanja = nabi_candidate_get_nth(candidate, 7);
 	break;
     case GDK_KP_Prior:
-	ch = nabi_candidate_get_nth(candidate, 8);
+	hanja = nabi_candidate_get_nth(candidate, 8);
 	break;
     case GDK_KP_Insert:
-	ch = nabi_candidate_get_nth(candidate, 9);
+	hanja = nabi_candidate_get_nth(candidate, 9);
 	break;
     default:
 	return FALSE;
     }
 
-    if (ch != 0) {
+    if (hanja != NULL) {
 	if (candidate->commit != NULL)
-	    candidate->commit(candidate, candidate->commit_data);
+	    candidate->commit(candidate, hanja, candidate->commit_data);
     }
     return TRUE;
 }
@@ -233,7 +237,8 @@ nabi_candidate_set_window_position(NabiCandidate *candidate)
 {
     Window root;
     Window child;
-    int x = 0, y = 0, width = 0, height = 0, border = 0, depth = 0;
+    int x = 0, y = 0;
+    unsigned int width = 0, height = 0, border = 0, depth = 0;
     int absx = 0;
     int absy = 0;
     int root_w, root_h, cand_w, cand_h;
@@ -272,22 +277,21 @@ static void
 nabi_candidate_update_list(NabiCandidate *candidate)
 {
     int i;
-    int len;
-    gchar buf[16];
     GtkTreeIter iter;
 
     gtk_list_store_clear(candidate->store);
     for (i = 0;
 	 i < candidate->n_per_page && candidate->first + i < candidate->n;
 	 i++) {
-	len = g_unichar_to_utf8(candidate->data[candidate->first + i]->ch,
-				buf);
-	buf[len] = '\0';
+	const Hanja* hanja = candidate->data[candidate->first + i];
+	const char* value = hanja_get_value(hanja);
+	const char* comment = hanja_get_comment(hanja);
+
 	gtk_list_store_append(candidate->store, &iter);
 	gtk_list_store_set(candidate->store, &iter,
 			   COLUMN_INDEX, (i + 1) % 10,
-			   COLUMN_CHARACTER, buf,
-			   COLUMN_COMMENT, candidate->data[candidate->first + i]->comment,
+			   COLUMN_CHARACTER, value,
+			   COLUMN_COMMENT, comment,
 			   -1);
     }
     nabi_candidate_set_window_position(candidate);
@@ -365,14 +369,14 @@ nabi_candidate_create_window(NabiCandidate *candidate)
 NabiCandidate*
 nabi_candidate_new(char *label_str,
 		   int n_per_page,
-		   const NabiCandidateItem **data,
+		   HanjaList *list,
+		   const Hanja **valid_list,
+		   int valid_list_length,
 		   Window parent,
 		   NabiCandidateCommitFunc commit,
 		   gpointer commit_data)
 {
-    int i, k, n;
     NabiCandidate *candidate;
-    const NabiCandidateItem **table;
 
     candidate = (NabiCandidate*)g_malloc(sizeof(NabiCandidate));
     candidate->first = 0;
@@ -386,24 +390,10 @@ nabi_candidate_new(char *label_str,
     candidate->treeview = NULL;
     candidate->commit = commit;
     candidate->commit_data = commit_data;
+    candidate->hanja_list = list;
 
-    for (n = 0; data[n] != 0; n++)
-	continue;
-
-    table = g_malloc(sizeof(NabiCandidateItem*) * n);
-    for (i = 0, k = 0; i < n; i++) {
-	if (nabi_server->check_charset) {
-	    if (nabi_server_is_valid_char(nabi_server, data[i]->ch)) {
-		table[k] = data[i];
-		k++;
-	    }
-	} else {
-	    table[k] = data[i];
-	    k++;
-	}
-    }
-    candidate->data = table;
-    candidate->n = k;
+    candidate->data = valid_list;
+    candidate->n = valid_list_length;
 
     candidate->store = gtk_list_store_new(NO_OF_COLUMNS,
 				    G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING);
@@ -480,16 +470,16 @@ nabi_candidate_next_page(NabiCandidate *candidate)
     nabi_candidate_update_cursor(candidate);
 }
 
-unsigned short int
+const Hanja*
 nabi_candidate_get_current(NabiCandidate *candidate)
 {
     if (candidate == NULL)
 	return 0;
 
-    return candidate->data[candidate->current]->ch;
+    return candidate->data[candidate->current];
 }
 
-unsigned short int
+const Hanja*
 nabi_candidate_get_nth(NabiCandidate *candidate, int n)
 {
     if (candidate == NULL)
@@ -500,7 +490,7 @@ nabi_candidate_get_nth(NabiCandidate *candidate, int n)
 	return 0;
 
     candidate->current = n;
-    return candidate->data[n]->ch;
+    return candidate->data[n];
 }
 
 void
@@ -509,29 +499,10 @@ nabi_candidate_delete(NabiCandidate *candidate)
     if (candidate == NULL)
 	return;
 
+    hanja_list_delete(candidate->hanja_list);
     gtk_grab_remove(candidate->window);
     gtk_widget_destroy(candidate->window);
     g_free(candidate->label);
     g_free(candidate->data);
     g_free(candidate);
 }
-
-NabiCandidateItem*
-nabi_candidate_item_new(unsigned short int ch, const gchar *comment)
-{
-    NabiCandidateItem *item;
-
-    item = g_new(NabiCandidateItem, 1);
-    item->ch = ch;
-    item->comment = g_strdup(comment);
-    return item;
-}
-
-void
-nabi_candidate_item_delete(NabiCandidateItem *item)
-{
-    if (item) {
-	g_free(item->comment);
-	g_free(item);
-    }
-}
