@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004 Choe Hwanjin
+ * Copyright (C) 2003,2004,2007 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,14 +22,10 @@
 #include <X11/Xlib.h>
 #include <gtk/gtk.h>
 
-typedef struct _NabiCandidate     NabiCandidate;
-typedef struct _NabiCandidateItem NabiCandidateItem;
-typedef void (*NabiCandidateCommitFunc)(NabiCandidate*, gpointer);
+#include <hangul.h>
 
-struct _NabiCandidateItem {
-    unsigned short int ch;
-    gchar *comment;
-};
+typedef struct _NabiCandidate     NabiCandidate;
+typedef void (*NabiCandidateCommitFunc)(NabiCandidate*, const Hanja*, gpointer);
 
 struct _NabiCandidate {
     GtkWidget *window;
@@ -37,18 +33,21 @@ struct _NabiCandidate {
     gchar *label;
     GtkListStore *store;
     GtkWidget *treeview;
-    const NabiCandidateItem **data;
+    const Hanja **data;
     NabiCandidateCommitFunc commit;
     gpointer commit_data;
     int first;
     int n;
     int n_per_page;
     int current;
+    HanjaList *hanja_list;
 };
 
 NabiCandidate*     nabi_candidate_new(char *label_str,
 		   	              int n_per_page,
-			              const NabiCandidateItem **data,
+			              HanjaList* list,
+			              const Hanja** valid_list,
+			              int valid_list_length,
 			              Window parent,
 				      NabiCandidateCommitFunc commit,
 				      gpointer commit_data);
@@ -58,12 +57,8 @@ void               nabi_candidate_prev_row(NabiCandidate *candidate);
 void               nabi_candidate_next_row(NabiCandidate *candidate);
 void               nabi_candidate_prev_page(NabiCandidate *candidate);
 void               nabi_candidate_next_page(NabiCandidate *candidate);
-unsigned short int nabi_candidate_get_current(NabiCandidate *candidate);
-unsigned short int nabi_candidate_get_nth(NabiCandidate *candidate, int n);
+const Hanja*       nabi_candidate_get_current(NabiCandidate *candidate);
+const Hanja*       nabi_candidate_get_nth(NabiCandidate *candidate, int n);
 void               nabi_candidate_delete(NabiCandidate *candidate);
 
-NabiCandidateItem* nabi_candidate_item_new(unsigned short int ch,
-					   const gchar *comment);
-void               nabi_candidate_item_delete(NabiCandidateItem *item);
-
 #endif /* _NABICANDIDATE_H_ */
