@@ -25,6 +25,7 @@
 #include <X11/Xlib.h>
 #include <glib.h>
 
+#include "debug.h"
 #include "gettext.h"
 #include "fontset.h"
 
@@ -36,14 +37,13 @@ static NabiFontSet*
 nabi_fontset_new(const char *name)
 {
     XFontSet xfontset;
-    XFontStruct **font_structs;
-    int i, nfonts, ascent, descent;
-    char **font_names;
+    XFontSetExtents* ext;
     NabiFontSet *fontset;
     char **missing_list;
     int    missing_list_count;
     char  *error_message;
 
+    nabi_log(4, "create fontset: %s\n", name);
     xfontset = XCreateFontSet(_display,
                               name,
                               &missing_list,
@@ -60,31 +60,24 @@ nabi_fontset_new(const char *name)
 				  &error_message);
 	g_free(name2);
 	if (missing_list_count > 0) {
-	    fprintf(stderr, _("Nabi: missing charset\n"));
-	    fprintf(stderr, _("Nabi: font: %s\n"), name);
+	    int i;
+	    nabi_log(1, "missing charset\n");
+	    nabi_log(1, "font: %s\n", name);
 	    for (i = 0; i < missing_list_count; i++)
-		fprintf(stderr, "  %s\n", missing_list[i]);
+		nabi_log(1, "  %s\n", missing_list[i]);
 	    XFreeStringList(missing_list);
 	    return NULL;
 	}
     }
 
-    /* get acsent and descent */
-    nfonts = XFontsOfFontSet(xfontset, &font_structs, &font_names);
-    /* find width, height */
-    for (i = 0, ascent = 1, descent = 0; i < nfonts; i++) {
-        if (ascent < font_structs[i]->ascent)
-            ascent = font_structs[i]->ascent;
-        if (descent < font_structs[i]->descent)
-            descent = font_structs[i]->descent;
-    }
+    ext = XExtentsOfFontSet(xfontset);
 
     fontset = g_malloc(sizeof(NabiFontSet));
     fontset->name = g_strdup(name);
     fontset->ref = 1;
     fontset->xfontset = xfontset;
-    fontset->ascent = ascent;
-    fontset->descent = descent;
+    fontset->ascent = ABS(ext->max_logical_extent.y);
+    fontset->descent = ext->max_logical_extent.height - fontset->ascent;
 
     g_hash_table_insert(fontset_hash, fontset->name, fontset);
     fontset_list = g_slist_prepend(fontset_list, fontset);
@@ -112,6 +105,7 @@ nabi_fontset_unref(NabiFontSet *fontset)
 	g_hash_table_remove(fontset_hash, fontset->name);
 	fontset_list = g_slist_remove(fontset_list, fontset);
 
+	nabi_log(4, "delete fontset: %s\n", fontset->name);
 	XFreeFontSet(_display, fontset->xfontset);
 	g_free(fontset->name);
 	g_free(fontset);
@@ -173,9 +167,8 @@ nabi_fontset_free_all(Display *display)
     _display = display;
 
     if (fontset_list != NULL) {
-	fprintf(stderr,
-		"Nabi: remaining fontsets will be freed,"
-		"this must be an error\n");
+	nabi_log(1, "remaining fontsets will be freed,"
+		 "this must be an error\n");
 	list = fontset_list;
 	while (list != NULL) {
 	    fontset = (NabiFontSet*)(list->data);
