@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004 Choe Hwanjin
+ * Copyright (C) 2003,2004,2007 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,6 +24,8 @@
 #include <glib.h>
 #include <gtk/gtk.h>
 
+#include <hangul.h>
+
 #include "../IMdkit/IMdkit.h"
 #include "../IMdkit/Xi18n.h"
 
@@ -31,22 +33,23 @@
 
 typedef struct _PreeditAttributes PreeditAttributes;
 typedef struct _StatusAttributes StatusAttributes;
-typedef struct _NabiIC NabiIC;
-typedef struct _NabiConnect NabiConnect;
+typedef struct _NabiIC         NabiIC;
+typedef struct _NabiConnection NabiConnection;
 
 typedef enum {
     NABI_INPUT_MODE_DIRECT,
     NABI_INPUT_MODE_COMPOSE
 } NabiInputMode;
 
-struct _NabiConnect {
-    CARD16               id;
-    NabiInputMode        mode;
-    GSList              *ic_list;
-    struct _NabiConnect *next;
+struct _NabiConnection {
+    CARD16         id;
+    NabiInputMode  mode;
+    GIConv         cd;
+    GSList*        ic_list;
 };
 
 struct _PreeditAttributes {
+    GString*        str;
     GdkWindow*      window;         /* where to draw the preedit string */
     int             width;          /* preedit area width */
     int             height;         /* preedit area height */
@@ -56,7 +59,8 @@ struct _PreeditAttributes {
     XRectangle      area_needed;    /* area needed */
 
     Colormap        cmap;           /* colormap */
-    GdkGC*          gc;             /* gc */
+    GdkGC*          normal_gc;      /* gc */
+    GdkGC*          hilight_gc;     /* gc */
     unsigned long   foreground;     /* foreground */
     unsigned long   background;     /* background */
 
@@ -88,7 +92,6 @@ struct _StatusAttributes {
 
 struct _NabiIC {
     CARD16              id;               /* ic id */
-    CARD16              connect_id;       /* connect id */
     INT32               input_style;      /* input style */
     Window              client_window;    /* client window */
     Window              focus_window;     /* focus window */
@@ -97,46 +100,35 @@ struct _NabiIC {
     StatusAttributes    status_attr;      /* status attributes */
     PreeditAttributes   preedit;          /* preedit attributes */
 
-    NabiConnect*        connect;
+    NabiConnection*     connection;
 
     /* hangul data */
     NabiInputMode       mode;
-    int                 index;            /* stack index */
-    wchar_t             stack[12];
-
-    int                 lindex;           /* choseong  index */
-    int                 vindex;           /* jungsoeng index */
-    int                 tindex;           /* jongseong index */
-    wchar_t             choseong[4];
-    wchar_t             jungseong[4];
-    wchar_t             jongseong[4];
+    HangulInputContext* hic;
+    gboolean            has_cho;
+    gboolean            has_jung;
+    gboolean            has_jong;
 
     /* hanja or symbol select window */
     NabiCandidate*	candidate;
-
-    struct _NabiIC*     next;
 };
 
-NabiConnect* nabi_connect_create(CARD16 id);
-void         nabi_connect_destroy(NabiConnect* connect);
-void         nabi_connect_add_ic(NabiConnect* connect, NabiIC *ic);
-void         nabi_connect_remove_ic(NabiConnect* connect, NabiIC *ic);
+NabiConnection* nabi_connection_create(CARD16 id, const char* encoding);
+void         nabi_connection_destroy(NabiConnection* conn);
+NabiIC*      nabi_connection_create_ic(NabiConnection* conn,
+				       IMChangeICStruct* data);
+void         nabi_connection_destroy_ic(NabiConnection* conn, NabiIC* ic);
 
-NabiIC* nabi_ic_create(IMChangeICStruct *data);
+
+NabiIC* nabi_ic_create(NabiConnection* conn, IMChangeICStruct *data);
 void    nabi_ic_destroy(NabiIC *ic);
 void    nabi_ic_real_destroy(NabiIC *ic);
 
 void    nabi_ic_set_values(NabiIC *ic, IMChangeICStruct *data);
 void    nabi_ic_get_values(NabiIC *ic, IMChangeICStruct *data);
 
-void    nabi_ic_reset(NabiIC *ic, IMResetICStruct *data);
-
-#define nabi_ic_is_empty(ic)        ((ic)->choseong[0]  == 0 &&     \
-                                     (ic)->jungseong[0] == 0 &&     \
-                                     (ic)->jongseong[0] == 0 )
-void    nabi_ic_push(NabiIC *ic, wchar_t ch);
-wchar_t nabi_ic_peek(NabiIC *ic);
-wchar_t nabi_ic_pop(NabiIC *ic);
+Bool    nabi_ic_is_empty(NabiIC *ic);
+CARD16  nabi_ic_get_id(NabiIC* ic);
 
 void    nabi_ic_mode_direct(NabiIC *ic);
 void    nabi_ic_mode_compose(NabiIC *ic);
@@ -144,7 +136,6 @@ void    nabi_ic_set_mode(NabiIC *ic, NabiInputMode mode);
 
 void    nabi_ic_preedit_start(NabiIC *ic);
 void    nabi_ic_preedit_done(NabiIC *ic);
-void    nabi_ic_preedit_insert(NabiIC *ic);
 void    nabi_ic_preedit_update(NabiIC *ic);
 void    nabi_ic_preedit_clear(NabiIC *ic);
 
@@ -155,8 +146,12 @@ void    nabi_ic_status_update(NabiIC *ic);
 Bool    nabi_ic_commit(NabiIC *ic);
 Bool    nabi_ic_commit_keyval(NabiIC *ic, wchar_t ch, KeySym keyval);
 
+Bool    nabi_ic_process_keyevent(NabiIC* ic, KeySym keysym, unsigned int state);
+void    nabi_ic_flush(NabiIC *ic);
+void    nabi_ic_reset(NabiIC *ic, IMResetICStruct *data);
+
 Bool    nabi_ic_popup_candidate_window(NabiIC *ic);
-void    nabi_ic_insert_candidate(NabiIC *ic, wchar_t ch);
+void    nabi_ic_insert_candidate(NabiIC *ic, const Hanja* hanja);
 
 #endif /* _NABIIC_H_ */
 /* vim: set ts=8 sw=4 sts=4 : */
