@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004 Choe Hwanjin
+ * Copyright (C) 2003,2004,2007 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,8 +21,11 @@
 #endif
 
 #include <stdio.h>
-#include <X11/Xlib.h>
+#include <stdlib.h>
 #include <langinfo.h>
+#include <signal.h>
+
+#include <X11/Xlib.h>
 #include <gdk/gdkx.h>
 #include <gtk/gtk.h>
 
@@ -30,10 +33,13 @@
 #include "server.h"
 #include "session.h"
 #include "nabi.h"
+#include "debug.h"
 
 NabiApplication* nabi = NULL;
 NabiServer* nabi_server = NULL;
 
+static void (*default_sigterm_handler)(int) = NULL;
+
 static void
 on_realize(GtkWidget *widget, gpointer data)
 {
@@ -57,6 +63,16 @@ nabi_x_error_handler(Display *display, XErrorEvent *error)
     return 0;
 }
 
+static void
+nabi_sigterm_handler(int no)
+{
+    nabi_app_save_config();
+    if (default_sigterm_handler != NULL)
+	default_sigterm_handler(no);
+    else
+	exit(0);
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -70,6 +86,8 @@ main(int argc, char *argv[])
 
     gtk_init(&argc, &argv);
 
+    nabi_log_set_device("stdout");
+
     nabi_app_new();
     nabi_app_init(&argc, &argv);
 
@@ -77,13 +95,15 @@ main(int argc, char *argv[])
     nabi_session_open(nabi->session_id);
 #endif
 
+    default_sigterm_handler = signal(SIGTERM, nabi_sigterm_handler);
+
     if (!nabi->status_only) {
 	char *xim_name;
 	/* we prefer command line option as default xim name */
-	if (nabi->optional_xim_name != NULL)
-	    xim_name = nabi->optional_xim_name;
-	else
+	if (nabi->xim_name != NULL)
 	    xim_name = nabi->xim_name;
+	else
+	    xim_name = nabi->config->xim_name;
 	nabi_server = nabi_server_new(xim_name);
 	nabi_server_init(nabi_server);
     }
