@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004 Choe Hwanjin
+ * Copyright (C) 2003,2004,2007 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,46 +22,27 @@
 #include <gtk/gtk.h>
 #include <X11/Xlib.h>
 
+#include "conf.h"
+
 typedef struct _NabiApplication NabiApplication;
 
 struct _NabiApplication {
-    gchar*	    xim_name;		/* xim server name from config file */
-    gchar*	    optional_xim_name;	/* xim server name from command line */
+    gchar*	    xim_name;
 
-    gint            x;
-    gint            y;
     GtkWidget*      main_window;
     gboolean	    status_only;
     gchar*	    session_id;
 
-    gchar*          theme;
     int             icon_size;
 
-    gchar*          keyboard_table_name;
-    gchar*          keyboard_table_dir;
-    gchar*          compose_table_name;
-    gchar*          compose_table_dir;
-    gchar*	    candidate_table_filename;
-
-    gchar*          trigger_keys;
-    gchar*          candidate_keys;
-
-    /* xim server option */
-    gboolean        dvorak;
-    gchar           *output_mode;
-
-    /* candidate options */
-    gchar*	    candidate_font;
-
-    /* preedit attribute */
-    gchar           *preedit_fg;
-    gchar           *preedit_bg;
-
     /* hangul status data */
     GdkWindow       *root_window;
     GdkAtom         mode_info_atom;
     GdkAtom         mode_info_type;
     Atom            mode_info_xatom;
+
+    /* config data */
+    NabiConfig*     config;
 };
 
 extern NabiApplication* nabi;
@@ -71,13 +52,9 @@ void nabi_app_init(int *argc, char ***argv);
 void nabi_app_setup_server(void);
 void nabi_app_quit(void);
 void nabi_app_free(void);
-void nabi_save_config_file(void);
+void nabi_app_save_config(void);
 void nabi_app_set_theme(const char *name);
-void nabi_app_set_dvorak(gboolean state);
-void nabi_app_set_keyboard(const char *name);
-void nabi_app_set_candidate_font(const char *font);
-void nabi_app_set_trigger_keys(char **keys);
-void nabi_app_set_candidate_keys(char **keys);
+void nabi_app_set_hangul_keyboard(const char *id);
 
 GtkWidget* nabi_app_create_main_widget(void);
 
