@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004 Choe Hwanjin
+ * Copyright (C) 2003,2004,2007 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -41,7 +41,8 @@ enum {
     N_COLS
 };
 
-static GtkWidget *keyboard_list_treeview = NULL;
+static NabiConfig* config = NULL;
+static GtkWidget *hangul_keyboard_list_combo = NULL;
 
 static GdkPixbuf *
 load_resized_icons_from_file(const gchar *filename, int size)
@@ -263,7 +264,7 @@ create_theme_page(void)
     g_signal_connect(G_OBJECT(selection), "changed",
 		     G_CALLBACK(on_icon_list_selection_changed), NULL);
 
-    path = search_text_in_model(model, THEMES_LIST_NAME, nabi->theme);
+    path = search_text_in_model(model, THEMES_LIST_NAME, config->theme);
     if (path) {
 	gtk_tree_view_set_cursor (GTK_TREE_VIEW(treeview), path, NULL, FALSE);
 	gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(treeview),
@@ -274,46 +275,30 @@ create_theme_page(void)
     return page;
 }
 
-static GtkTreeModel*
-get_keyboard_list(void)
-{
-    GList *list;
-    GtkListStore *store;
-    GtkTreeIter iter;
-
-    store = gtk_list_store_new(1, G_TYPE_STRING);
-
-    list = nabi_server->keyboard_tables;
-    while (list != NULL) {
-	NabiKeyboardTable *table = (NabiKeyboardTable*)list->data;
-	gtk_list_store_append(store, &iter);
-	gtk_list_store_set (store, &iter, 0, table->name, -1);
-	list = g_list_next(list);
-    }
-
-    return GTK_TREE_MODEL(store);
-}
-
 static void
-on_keyboard_list_selection_changed(GtkTreeSelection *selection, gpointer data)
+on_hangul_keyboard_changed(GtkComboBox *widget, gpointer data)
 {
-    GtkTreeModel *model;
-    GtkTreeIter iter;
-    gchar *name = NULL;
+    int i;
+    const NabiHangulKeyboard* keyboards;
 
-    if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
-	gtk_tree_model_get(model, &iter, 0, &name, -1);
-	nabi_app_set_keyboard(name);
-	g_free(name);
+    i = gtk_combo_box_get_active(widget);
+    keyboards = nabi_server_get_hangul_keyboard_list(nabi_server);
+
+    if (keyboards[i].id != NULL) {
+	nabi_app_set_hangul_keyboard(keyboards[i].id);
     }
 }
 
 static void
-on_dvorak_button_toggled(GtkToggleButton *button, gpointer data)
+on_latin_keyboard_changed(GtkComboBox *widget, gpointer data)
 {
-    gboolean state;
-    state = gtk_toggle_button_get_active(button);
-    nabi_app_set_dvorak(state);
+    int i = gtk_combo_box_get_active(widget);
+    NabiKeyboardLayout* item = g_list_nth_data(nabi_server->layouts, i);
+    if (item != NULL) {
+	g_free(config->latin_keyboard);
+	config->latin_keyboard = g_strdup(item->name);
+	nabi_server_set_keyboard_layout(nabi_server, item->name);
+    }
 }
 
 static GtkWidget*
@@ -324,20 +309,17 @@ create_keyboard_page(void)
     GtkWidget *vbox1;
     GtkWidget *vbox2;
     GtkWidget *label;
-    GtkWidget *scrolledwindow;
-    GtkWidget *treeview;
-    GtkTreeViewColumn *column;
-    GtkTreeSelection *selection;
-    GtkCellRenderer *renderer;
-    GtkTreeModel *model;
-    GtkTreePath *path;
-    GtkWidget *radio_button;
+    GtkWidget *combo_box;
+    GtkSizeGroup* size_group;
+    GList* list;
+    int i;
+    const NabiHangulKeyboard* keyboards;
 
     page = gtk_vbox_new(FALSE, 12);
     gtk_container_set_border_width(GTK_CONTAINER(page), 12);
 
     vbox1 = gtk_vbox_new(FALSE, 6);
-    gtk_box_pack_start(GTK_BOX(page), vbox1, TRUE, TRUE, 0);
+    gtk_box_pack_start(GTK_BOX(page), vbox1, FALSE, TRUE, 0);
 
     label = gtk_label_new("");
     gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
@@ -353,34 +335,24 @@ create_keyboard_page(void)
     gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);
 
     vbox2 = gtk_vbox_new(FALSE, 0);
-    gtk_box_pack_start(GTK_BOX(hbox), vbox2, TRUE, TRUE, 0);
+    gtk_box_pack_start(GTK_BOX(hbox), vbox2, FALSE, FALSE, 0);
 
-    scrolledwindow = gtk_scrolled_window_new(NULL, NULL);
-    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwindow),
-				   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
-    gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolledwindow),
-					GTK_SHADOW_IN);
-    gtk_container_set_border_width(GTK_CONTAINER(scrolledwindow), 0);
-    gtk_box_pack_start(GTK_BOX(vbox2), scrolledwindow, TRUE, TRUE, 0);
+    size_group = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
 
-    model = get_keyboard_list();
-    treeview = gtk_tree_view_new_with_model(model);
-    g_object_unref(G_OBJECT(model));
-    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), FALSE);
-    gtk_container_add(GTK_CONTAINER(scrolledwindow), treeview);
-
-    column = gtk_tree_view_column_new();
-    gtk_tree_view_column_set_title(column, _("Hangul Keyboard"));
-    renderer = gtk_cell_renderer_text_new();
-    gtk_tree_view_column_pack_start(column, renderer, FALSE);
-    gtk_tree_view_column_add_attribute(column, renderer, "text", 0);
-    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
-
-    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
-    gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
-
-    g_signal_connect(G_OBJECT(selection), "changed",
-		     G_CALLBACK(on_keyboard_list_selection_changed), NULL);
+    combo_box = gtk_combo_box_new_text();
+    keyboards = nabi_server_get_hangul_keyboard_list(nabi_server);
+    for (i = 0; keyboards[i].name != NULL; i++) {
+	gtk_combo_box_append_text(GTK_COMBO_BOX(combo_box),
+				  _(keyboards[i].name));
+	if (strcmp(config->hangul_keyboard, keyboards[i].id) == 0) {
+	    gtk_combo_box_set_active(GTK_COMBO_BOX(combo_box), i);
+	}
+    }
+    gtk_box_pack_start(GTK_BOX(vbox2), combo_box, FALSE, FALSE, 0);
+    gtk_size_group_add_widget(size_group, combo_box);
+    g_signal_connect(G_OBJECT(combo_box), "changed",
+		     G_CALLBACK(on_hangul_keyboard_changed), NULL);
+    hangul_keyboard_list_combo = combo_box;
 
     vbox1 = gtk_vbox_new(FALSE, 6);
     gtk_box_pack_start(GTK_BOX(page), vbox1, FALSE, TRUE, 0);
@@ -401,28 +373,27 @@ create_keyboard_page(void)
     vbox2 = gtk_vbox_new(FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), vbox2, FALSE, TRUE, 0);
 
-    radio_button = gtk_radio_button_new_with_label(NULL,
-					_("Querty keyboard"));
-    gtk_box_pack_start(GTK_BOX(vbox2), radio_button, FALSE, TRUE, 0);
-    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio_button),!nabi->dvorak);
-
-    radio_button = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(radio_button),
-							_("Dvorak keyboard"));
-    gtk_box_pack_start(GTK_BOX(vbox2), radio_button, FALSE, TRUE, 0);
-    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio_button), nabi->dvorak);
-    g_signal_connect(G_OBJECT(radio_button), "toggled",
-		     G_CALLBACK(on_dvorak_button_toggled), NULL);
-
-    /* hilight current selected keyboard table */
-    path = search_text_in_model(model, 0, nabi->keyboard_table_name);
-    if (path) {
-	gtk_tree_view_set_cursor (GTK_TREE_VIEW(treeview), path, NULL, FALSE);
-	gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(treeview),
-				    path, NULL, TRUE, 0.5, 0.0);
-	gtk_tree_path_free(path);
+    /* latin keyboard */
+    combo_box = gtk_combo_box_new_text();
+    i = 0;
+    list = nabi_server->layouts;
+    while (list != NULL) {
+	NabiKeyboardLayout* layout = list->data;
+	if (layout != NULL) {
+	    gtk_combo_box_append_text(GTK_COMBO_BOX(combo_box), layout->name);
+	    if (strcmp(config->latin_keyboard, layout->name) == 0) {
+		gtk_combo_box_set_active(GTK_COMBO_BOX(combo_box), i);
+	    }
+	}
+	list = g_list_next(list);
+	i++;
     }
+    gtk_box_pack_start(GTK_BOX(vbox2), combo_box, FALSE, FALSE, 0);
+    gtk_size_group_add_widget(size_group, combo_box);
+    g_signal_connect(G_OBJECT(combo_box), "changed",
+		     G_CALLBACK(on_latin_keyboard_changed), NULL);
 
-    keyboard_list_treeview = treeview;
+    g_object_unref(G_OBJECT(size_group));
 
     return page;
 }
@@ -468,12 +439,16 @@ update_trigger_keys_setting(GtkTreeModel *model)
     while (ret) {
 	char *key = NULL;
 	gtk_tree_model_get(model, &iter, 0, &key, -1);
-	keys[n++] = key;
+	if (key != NULL)
+	    keys[n++] = key;
 	ret = gtk_tree_model_iter_next(model, &iter);
     }
     keys[n] = NULL;
 
-    nabi_app_set_trigger_keys(keys);
+    g_free(config->trigger_keys);
+    config->trigger_keys = g_strjoinv(",", keys);
+    nabi_server_set_trigger_keys(nabi_server, keys);
+
     g_strfreev(keys);
 }
 
@@ -560,12 +535,15 @@ update_candidate_keys_setting(GtkTreeModel *model)
     while (ret) {
 	char *key = NULL;
 	gtk_tree_model_get(model, &iter, 0, &key, -1);
-	keys[n++] = key;
+	if (key != NULL)
+	    keys[n++] = key;
 	ret = gtk_tree_model_iter_next(model, &iter);
     }
     keys[n] = NULL;
 
-    nabi_app_set_candidate_keys(keys);
+    g_free(config->candidate_keys);
+    config->candidate_keys = g_strjoinv(",", keys);
+    nabi_server_set_candidate_keys(nabi_server, keys);
     g_strfreev(keys);
 }
 
@@ -663,7 +641,7 @@ create_hangul_page(void)
     gtk_container_set_border_width(GTK_CONTAINER(scrolledwindow), 0);
     gtk_box_pack_start(GTK_BOX(vbox2), scrolledwindow, TRUE, TRUE, 0);
 
-    model = get_key_list_store(nabi->trigger_keys);
+    model = get_key_list_store(config->trigger_keys);
     treeview = gtk_tree_view_new_with_model(model);
     g_object_unref(G_OBJECT(model));
     gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), FALSE);
@@ -717,7 +695,7 @@ on_candidate_font_button_clicked(GtkWidget *button, gpointer data)
 
     dialog = gtk_font_selection_dialog_new(_("Select hanja font"));
     gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(dialog),
-					    nabi->candidate_font);
+					    config->candidate_font);
     gtk_font_selection_dialog_set_preview_text(GTK_FONT_SELECTION_DIALOG(dialog),
 					       "訓民正音 훈민정음");
     gtk_widget_show_all(dialog);
@@ -727,7 +705,9 @@ on_candidate_font_button_clicked(GtkWidget *button, gpointer data)
 
     if (result == GTK_RESPONSE_OK) {
 	char *font = gtk_font_selection_dialog_get_font_name(GTK_FONT_SELECTION_DIALOG(dialog));
-	nabi_app_set_candidate_font(font);
+	g_free(config->candidate_font);
+	config->candidate_font = g_strdup(font);
+	nabi_server_set_candidate_font(nabi_server, font);
 	gtk_button_set_label(GTK_BUTTON(button), font);
     }
     gtk_widget_destroy(dialog);
@@ -779,7 +759,7 @@ create_candidate_page(void)
     hbox2 = gtk_hbox_new(FALSE, 0);
     gtk_box_pack_start(GTK_BOX(vbox2), hbox2, FALSE, TRUE, 0);
 
-    button = gtk_button_new_with_label(nabi->candidate_font);
+    button = gtk_button_new_with_label(config->candidate_font);
     gtk_box_pack_start(GTK_BOX(hbox2), button, FALSE, TRUE, 0);
     g_signal_connect(G_OBJECT(button), "clicked",
 		     G_CALLBACK(on_candidate_font_button_clicked), NULL);
@@ -812,7 +792,7 @@ create_candidate_page(void)
     gtk_container_set_border_width(GTK_CONTAINER(scrolledwindow), 0);
     gtk_box_pack_start(GTK_BOX(vbox2), scrolledwindow, TRUE, TRUE, 0);
 
-    model = get_key_list_store(nabi->candidate_keys);
+    model = get_key_list_store(config->candidate_keys);
     treeview = gtk_tree_view_new_with_model(model);
     g_object_unref(G_OBJECT(model));
     gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), FALSE);
@@ -845,19 +825,132 @@ create_candidate_page(void)
     return page;
 }
 
+static void
+on_xim_name_changed(GtkEntry* entry, gpointer data)
+{
+    const char* name = gtk_entry_get_text(GTK_ENTRY(entry));
+
+    g_free(config->xim_name);
+    config->xim_name = g_strdup(name);
+    nabi_server_set_xim_name(nabi_server, name);
+}
+
+static void
+on_event_flow_button_toggled(GtkToggleButton *button, gpointer data)
+{
+    gboolean flag = gtk_toggle_button_get_active(button);
+
+    config->use_dynamic_event_flow = flag;
+    nabi_server_set_dynamic_event_flow(nabi_server, flag);
+}
+
+static void
+on_commit_by_word_button_toggled(GtkToggleButton *button, gpointer data)
+{
+    gboolean flag = gtk_toggle_button_get_active(button);
+
+    config->commit_by_word = flag;
+    nabi_server_set_commit_by_word(nabi_server, flag);
+}
+
+static void
+on_auto_reorder_button_toggled(GtkToggleButton *button, gpointer data)
+{
+    gboolean flag = gtk_toggle_button_get_active(button);
+
+    config->auto_reorder = flag;
+    nabi_server_set_auto_reorder(nabi_server, flag);
+}
+
 static GtkWidget*
 create_advanced_page(void)
 {
     GtkWidget *page;
+    GtkWidget *vbox1;
+    GtkWidget *vbox2;
+    GtkWidget *hbox;
+    GtkWidget *hbox2;
+    GtkWidget *label;
+    GtkWidget *button;
+    GtkWidget *entry;
+
+    page = gtk_vbox_new(FALSE, 12);
+    gtk_container_set_border_width(GTK_CONTAINER(page), 12);
+
+    vbox1 = gtk_vbox_new(FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(page), vbox1, FALSE, TRUE, 0);
+
+    hbox = gtk_hbox_new(FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(vbox1), hbox, FALSE, TRUE, 0);
+
+    label = gtk_label_new("");
+    gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
+    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
+    gtk_label_set_markup(GTK_LABEL(label),
+			 _("<span weight=\"bold\">XIM</span>"));
+    gtk_box_pack_start(GTK_BOX(vbox1), label, FALSE, TRUE, 0);
+
+    hbox = gtk_hbox_new(FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(vbox1), hbox, FALSE, TRUE, 6);
+
+    label = gtk_label_new("    ");
+    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);
+
+    vbox2 = gtk_vbox_new(FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(hbox), vbox2, FALSE, TRUE, 0);
+
+    hbox2 = gtk_hbox_new(FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(vbox2), hbox2, FALSE, FALSE, 0);
+
+    label = gtk_label_new(_("XIM name:"));
+    gtk_box_pack_start(GTK_BOX(hbox2), label, FALSE, FALSE, 0);
+
+    entry = gtk_entry_new();
+    gtk_entry_set_width_chars(GTK_ENTRY(entry), 16);
+    gtk_entry_set_text(GTK_ENTRY(entry), config->xim_name);
+    gtk_box_pack_start(GTK_BOX(hbox2), entry, FALSE, FALSE, 6);
+    g_signal_connect(G_OBJECT(entry), "changed",
+		     G_CALLBACK(on_xim_name_changed), NULL);
+
+    hbox2 = gtk_hbox_new(FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(vbox2), hbox2, FALSE, FALSE, 0);
+
+    button = gtk_check_button_new_with_label(_("Use dynamic event flow"));
+    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
+			         config->use_dynamic_event_flow);
+    gtk_box_pack_start(GTK_BOX(hbox2), button, FALSE, FALSE, 0);
+    g_signal_connect(G_OBJECT(button), "toggled",
+		     G_CALLBACK(on_event_flow_button_toggled), NULL);
+
+    hbox2 = gtk_hbox_new(FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(vbox2), hbox2, FALSE, FALSE, 0);
+
+    button = gtk_check_button_new_with_label(_("Commit by word"));
+    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
+			         config->commit_by_word);
+    gtk_box_pack_start(GTK_BOX(hbox2), button, FALSE, FALSE, 0);
+    g_signal_connect(G_OBJECT(button), "toggled",
+		     G_CALLBACK(on_commit_by_word_button_toggled), NULL);
+
+    hbox2 = gtk_hbox_new(FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(vbox2), hbox2, FALSE, FALSE, 0);
+
+    button = gtk_check_button_new_with_label(_("Automatic reordering"));
+    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
+			         config->auto_reorder);
+    gtk_box_pack_start(GTK_BOX(hbox2), button, FALSE, FALSE, 0);
+    g_signal_connect(G_OBJECT(button), "toggled",
+		     G_CALLBACK(on_auto_reorder_button_toggled), NULL);
 
-    page = gtk_label_new("Advanced");
     return page;
 }
 
 static void
 on_preference_destroy(GtkWidget *dialog, gpointer data)
 {
-    keyboard_list_treeview = NULL;
+    nabi_app_save_config();
+
+    hangul_keyboard_list_combo = NULL;
 }
 
 GtkWidget*
@@ -870,6 +963,8 @@ preference_window_create(void)
     GtkWidget *child;
     GList *list;
 
+    config = nabi->config;
+
     dialog = gtk_dialog_new_with_buttons(_("Nabi Preferences"),
 					 NULL,
 					 GTK_DIALOG_MODAL | GTK_DIALOG_NO_SEPARATOR,
@@ -904,7 +999,7 @@ preference_window_create(void)
     /* advanced */
     label = gtk_label_new(_("Advanced"));
     child = create_advanced_page();
-    //gtk_notebook_append_page(GTK_NOTEBOOK(notebook), child, label);
+    gtk_notebook_append_page(GTK_NOTEBOOK(notebook), child, label);
 
     vbox = GTK_DIALOG(dialog)->vbox;
     gtk_box_pack_start(GTK_BOX(vbox), notebook, TRUE, TRUE, 0);
@@ -932,17 +1027,14 @@ preference_window_create(void)
 void
 preference_window_update(void)
 {
-    /* keyboard list update */
-    if (keyboard_list_treeview != NULL) {
-	GtkTreeModel *model;
-	GtkTreePath *path;
-	model = gtk_tree_view_get_model(GTK_TREE_VIEW(keyboard_list_treeview));
-	if (model != NULL) {
-	    path = search_text_in_model(model, 0, nabi->keyboard_table_name);
-	    if (path) {
-		gtk_tree_view_set_cursor (GTK_TREE_VIEW(keyboard_list_treeview),
-					  path, NULL, FALSE);
-		gtk_tree_path_free(path);
+    /* hangul keyboard list update */
+    if (hangul_keyboard_list_combo != NULL) {
+	int i;
+	const NabiHangulKeyboard* keyboards;
+	keyboards = nabi_server_get_hangul_keyboard_list(nabi_server);
+	for (i = 0; keyboards[i].name != NULL; i++) {
+	    if (strcmp(config->hangul_keyboard, keyboards[i].id) == 0) {
+		gtk_combo_box_set_active(GTK_COMBO_BOX(hangul_keyboard_list_combo), i);
 	    }
 	}
     }
