@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004 Choe Hwanjin
+ * Copyright (C) 2003,2004,2007 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,28 +27,42 @@
 #include <dirent.h>
 #include <locale.h>
 #include <time.h>
+
 #include <X11/Xlib.h>
 #include <X11/keysym.h>
+
 #include <glib.h>
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
 
+#include "debug.h"
 #include "gettext.h"
 #include "server.h"
 #include "fontset.h"
 #include "hangul.h"
 
-#include "keyboard.h"
+#define DEFAULT_IC_TABLE_SIZE	64
 
-#define DEFAULT_IC_TABLE_SIZE	512
+struct KeySymPair {
+    KeySym key;
+    KeySym value;
+};
 
-/* from automata.c */
-Bool nabi_automata_2(NabiIC* ic, KeySym keyval, unsigned int state);
-Bool nabi_automata_3(NabiIC* ic, KeySym keyval, unsigned int state);
+static NabiHangulKeyboard hangul_keyboard_list[] = {
+    { "2",  N_("2 set") },
+    { "32", N_("3 set with 2 set layout") },
+    { "3f", N_("3 set final") },
+    { "39", N_("3 set 390") },
+    { "3s", N_("3 set no-shift") },
+    { "3y", N_("3 set yetguel") },
+    { NULL, NULL }
+};
 
 /* from handler.c */
 Bool nabi_handler(XIMS ims, IMProtocol *call_data);
 
+static void nabi_server_delete_layouts(NabiServer* server);
+
 long nabi_filter_mask = KeyPressMask;
 
 /* Supported Inputstyles */
@@ -67,6 +81,14 @@ static XIMEncoding nabi_encodings[] = {
 };
 
 static char *nabi_locales[] = {
+    "ko_KR.UTF-8",
+    "ko_KR.utf8",
+    "ko_KR.eucKR",
+    "ko_KR.euckr",
+    "ko.UTF-8",
+    "ko.utf8",
+    "ko.eucKR",
+    "ko.euckr",
     "ko",
     NULL,
     NULL
@@ -96,32 +118,31 @@ nabi_server_new(const char *name)
     server->candidate_keys.count_keys = 0;
     server->candidate_keys.keylist = NULL;
 
-    /* connect list */
-    server->n_connected = 0;
-    server->connect_list = NULL;
+    /* connection list */
+    server->connections = NULL;
 
     /* init IC table */
+    server->last_icid = 0;
+    server->freed_icid = NULL;
     server->ic_table = g_new(NabiIC*, DEFAULT_IC_TABLE_SIZE);
     server->ic_table_size = DEFAULT_IC_TABLE_SIZE;
     for (i = 0; i < server->ic_table_size; i++)
 	server->ic_table[i] = NULL;
-    server->ic_freed = NULL;
 
     /* hangul data */
-    server->keyboard_tables = NULL;
-    server->keyboard_table = NULL;
-    server->compose_tables = NULL;
-    server->compose_table = NULL;
-    server->automata = NULL;
+    server->layouts = NULL;
+    server->layout = NULL;
+    server->hangul_keyboard = NULL;
+    server->hangul_keyboard_list = hangul_keyboard_list;
 
     server->dynamic_event_flow = True;
     server->global_input_mode = True;
-    server->dvorak = False;
+    server->commit_by_word = False;
+    server->auto_reorder = TRUE;
     server->input_mode = NABI_INPUT_MODE_DIRECT;
 
-    /* hangul converter */
-    server->check_charset = False;
-    server->converter = (GIConv)(-1);
+    /* hanja */
+    server->hanja_table = NULL;
 
     /* options */
     server->show_status = False;
@@ -147,9 +168,8 @@ nabi_server_new(const char *name)
 void
 nabi_server_destroy(NabiServer *server)
 {
-    int i, j;
-    GList *list;
-    NabiConnect *connect;
+    int i;
+    GSList *item;
 
     if (server == NULL)
 	return;
@@ -157,17 +177,21 @@ nabi_server_destroy(NabiServer *server)
     if (server->gc != NULL)
 	g_object_unref(G_OBJECT(server->gc));
 
-    /* destroy remaining connect */
-    while (server->connect_list != NULL) {
-	connect = server->connect_list;
-	server->connect_list = server->connect_list->next;
-	printf("remove connect id: 0x%x\n", connect->id);
-	nabi_connect_destroy(connect);
+    /* destroy remaining connections */
+    item = server->connections;
+    while (item != NULL) {
+	if (item->data != NULL) {
+	    NabiConnection* conn = (NabiConnection*)item->data;
+	    printf("remove connect id: 0x%x\n", conn->id);
+	    nabi_connection_destroy(conn);
+	}
+	item = g_slist_next(item);
     }
 
     /* destroy remaining input contexts */
     for (i = 0; i < server->ic_table_size; i++) {
-	nabi_ic_real_destroy(server->ic_table[i]);
+	if (server->ic_table[i] != NULL)
+	    nabi_ic_destroy(server->ic_table[i]);
 	server->ic_table[i] = NULL;
     }
     g_free(server->ic_table);
@@ -175,31 +199,15 @@ nabi_server_destroy(NabiServer *server)
     /* free remaining fontsets */
     nabi_fontset_free_all(server->display);
 
-    /* delete keyboard table */
-    for (list = server->keyboard_tables; list != NULL; list = list->next) {
-	g_free(list->data);
-	list->data = NULL;
-    }
-    g_list_free(server->keyboard_tables);
+    /* keyboard */
+    nabi_server_delete_layouts(server);
 
-    /* delete compose table */
-    for (list = server->compose_tables; list != NULL; list = list->next) {
-	NabiComposeTable *table = (NabiComposeTable *)list->data;
-	g_free(table->name);
-	g_free(table->items);
-	g_free(table);
-	list->data = NULL;
-    }
-    g_list_free(server->compose_tables);
+    if (server->hangul_keyboard != NULL)
+	g_free(server->hangul_keyboard);
 
-    /* delete candidate table */
-    for (i = 0; i < server->candidate_table_size; i++) {
-	for (j = 0;  server->candidate_table[i][j] != NULL; j++) {
-	    nabi_candidate_item_delete(server->candidate_table[i][j]);
-	}
-	g_free(server->candidate_table[i]);
-    }
-    g_free(server->candidate_table);
+    /* delete hanja table */
+    if (server->hanja_table != NULL)
+	hanja_table_delete(server->hanja_table);
 
     g_free(server->trigger_keys.keylist);
     g_free(server->candidate_keys.keylist);
@@ -209,92 +217,12 @@ nabi_server_destroy(NabiServer *server)
 }
 
 void
-nabi_server_set_keyboard_table(NabiServer *server,
-			       const char *name)
+nabi_server_set_hangul_keyboard(NabiServer *server, const char *id)
 {
-    GList *list;
-    NabiKeyboardTable *default_table;
+    if (server->hangul_keyboard != NULL)
+	g_free(server->hangul_keyboard);
 
-    if (server == NULL)
-	return;
-
-    if (server->keyboard_tables == NULL)
-	default_table = &nabi_keyboard_table_internal;
-    else
-	default_table = server->keyboard_tables->data;
-
-    if (name == NULL) {
-	server->keyboard_table = default_table;
-	if (default_table->type == NABI_KEYBOARD_2SET)
-	    server->automata = nabi_automata_2;
-	else
-	    server->automata = nabi_automata_3;
-	return;
-    }
-
-    for (list = server->keyboard_tables; list != NULL; list = list->next) {
-	NabiKeyboardTable *table = (NabiKeyboardTable*)list->data;
-	if (strcmp(table->name, name) == 0) {
-	    server->keyboard_table = table;
-	    /* set keyboard */
-	    if (table->type == NABI_KEYBOARD_2SET)
-		server->automata = nabi_automata_2;
-	    else
-		server->automata = nabi_automata_3;
-	    return;
-	}
-    }
-
-    fprintf(stderr, "Nabi: Cant find keyboard table: %s\n", name);
-    server->keyboard_table = default_table;
-    if (default_table->type == NABI_KEYBOARD_2SET)
-	server->automata = nabi_automata_2;
-    else
-	server->automata = nabi_automata_3;
-}
-
-void
-nabi_server_set_compose_table(NabiServer *server,
-			      const char *name)
-{
-    GList *list;
-    NabiComposeTable *default_table;
-
-    if (server == NULL)
-	return;
-
-    if (server->compose_tables == NULL)
-	default_table = &nabi_compose_table_internal;
-    else
-	default_table = server->compose_tables->data;
-
-    if (name == NULL) {
-	server->compose_table = default_table;
-	return;
-    }
-
-    for (list = server->compose_tables; list != NULL; list = list->next) {
-	NabiComposeTable *table = (NabiComposeTable*)list->data;
-	if (strcmp(table->name, name) == 0) {
-	    server->compose_table = table;
-	    return;
-	}
-    }
-
-    fprintf(stderr, "Nabi: Cant find compose table: %s\n", name);
-    server->compose_table = default_table;
-}
-
-void
-nabi_server_set_automata(NabiServer *server, NabiKeyboardType type)
-{
-    if (server == NULL)
-	return;
-
-    if (type == NABI_KEYBOARD_2SET)
-	server->automata = nabi_automata_2;
-    else
-	server->automata = nabi_automata_3;
+    server->hangul_keyboard = g_strdup(id);
 }
 
 void
@@ -306,26 +234,16 @@ nabi_server_set_mode_info_cb(NabiServer *server, NabiModeInfoCallback func)
     server->mode_info_cb = func;
 }
 
-void
-nabi_server_set_dvorak(NabiServer *server, Bool flag)
-{
-    if (server == NULL)
-	return;
-
-    server->dvorak = flag;
-}
-
 void
 nabi_server_set_output_mode(NabiServer *server, NabiOutputMode mode)
 {
     if (server == NULL)
 	return;
 
-    if (mode == NABI_OUTPUT_SYLLABLE)
+    if (mode == NABI_OUTPUT_SYLLABLE) {
+	server->output_mode = mode;
+    } else  {
 	server->output_mode = mode;
-    else  {
-	if (!server->check_charset)
-	    server->output_mode = mode;
     }
 }
 
@@ -361,7 +279,7 @@ xim_trigger_keys_create(char **keys, int n)
 	    if (strcmp("Shift", list[j]) == 0) {
 		keylist[i].modifier |= ShiftMask;
 		keylist[i].modifier_mask |= ShiftMask;
-	    } else if (strcmp("Ctrl", list[j]) == 0) {
+	    } else if (strcmp("Control", list[j]) == 0) {
 		keylist[i].modifier |= ControlMask;
 		keylist[i].modifier_mask |= ControlMask;
 	    } else if (strcmp("Alt", list[j]) == 0) {
@@ -431,25 +349,10 @@ nabi_server_init(NabiServer *server)
     nabi_server_set_trigger_keys(server, trigger_keys);
     nabi_server_set_candidate_keys(server, candidate_keys);
 
-    server->automata = nabi_automata_2;
     server->output_mode = NABI_OUTPUT_SYLLABLE;
 
-    server->candidate_table_size = 0;
-    server->candidate_table = NULL;
-
     /* check locale encoding */
-    server->check_charset = !g_get_charset(&charset);
-    if (server->check_charset) {
-	/* not utf8 encoding */
-	fprintf(stderr, "Nabi: Using charset: %s\n", charset);
-	server->converter = g_iconv_open(charset, "UTF-8");
-	if ((GIConv)server->converter == (GIConv)(-1)) {
-	    server->check_charset = False;
-	    fprintf(stderr,
-		    "Nabi: g_iconv_open error: %s\n"
-		    "We does not check charset\n", charset);
-	}
-    } else {
+    if (g_get_charset(&charset)) {
 	/* utf8 encoding */
 	/* We add current locale to nabi support  locales array,
 	 * so whatever the locale string is, if its encoding is utf8,
@@ -459,22 +362,73 @@ nabi_server_init(NabiServer *server)
 	    continue;
 	server->locales[i] = setlocale(LC_CTYPE, NULL);
     }
+
+    server->hanja_table = hanja_table_load(NULL);
+}
+
+NabiIC*
+nabi_server_alloc_ic(NabiServer* server)
+{
+    CARD16 icid = 0;
+    NabiIC* ic;
+
+    if (server->freed_icid == NULL) {
+	server->last_icid++;
+	if (server->last_icid == 0)
+	    server->last_icid++;  /* icid에 0은 사용하지 않는다 */
+
+	if (server->last_icid >= server->ic_table_size) {
+	    gint new_size = server->ic_table_size;
+	    while (new_size <= server->last_icid) {
+		new_size += DEFAULT_IC_TABLE_SIZE / 2;
+	    }
+	    server->ic_table = g_renew(NabiIC*, server->ic_table, new_size);
+	}
+
+	icid = server->last_icid;
+    } else {
+	/* pick from ic_freed */
+	icid = (CARD16)GPOINTER_TO_UINT(server->freed_icid->data);
+	server->freed_icid = g_slist_remove(server->freed_icid, 
+					    GUINT_TO_POINTER((guint)icid));
+    }
+
+    ic = g_new(NabiIC, 1);
+    ic->id = icid;
+
+    server->ic_table[ic->id] = ic;
+
+    return ic;
 }
 
 void
-nabi_server_ic_table_expand(NabiServer *server)
+nabi_server_dealloc_ic(NabiServer* server, NabiIC* ic)
+{
+    server->freed_icid = g_slist_append(server->freed_icid,
+					GUINT_TO_POINTER((guint)ic->id));
+    server->ic_table[ic->id] = NULL;
+
+    g_free(ic);
+}
+
+gboolean
+nabi_server_is_valid_ic(NabiServer* server, NabiIC* ic)
 {
     int i;
-    int old_size;
+    for (i = 0; i < server->ic_table_size; i++) {
+	if (server->ic_table[i] == ic)
+	    return TRUE;
+    }
+    return FALSE;
+}
 
-    if (server == NULL)
-	return;
-
-    old_size = server->ic_table_size;
-    server->ic_table_size = server->ic_table_size * 2;
-    server->ic_table = g_renew(NabiIC*, server->ic_table,server->ic_table_size);
-    for (i = old_size; i < server->ic_table_size; i++)
-	server->ic_table[i] = NULL;
+NabiIC*
+nabi_server_get_ic(NabiServer *server, CARD16 icid)
+{
+    if (icid > 0 && icid < server->ic_table_size)
+	return server->ic_table[icid];
+    else
+	return NULL;
 }
 
 static Bool
@@ -571,7 +525,7 @@ nabi_server_start(NabiServer *server, GtkWidget *widget)
     gdk_gc_set_foreground(server->gc, &(server->preedit_fg));
     gdk_gc_set_background(server->gc, &(server->preedit_bg));
 
-    fprintf(stderr, "Nabi: xim server started\n");
+    nabi_log(1, "xim server started\n");
 
     return 0;
 }
@@ -582,80 +536,59 @@ nabi_server_stop(NabiServer *server)
     if (server == NULL)
 	return 0;
 
-    if ((GIConv)(server->converter) != (GIConv)(-1)) {
-	g_iconv_close(server->converter);
-	server->converter = (GIConv)(-1);
-    }
-
     if (server->xims != NULL) {
 	IMCloseIM(server->xims);
 	server->xims = NULL;
     }
-    fprintf(stderr, "Nabi: xim server stoped\n");
+    nabi_log(1, "xim server stoped\n");
 
     return 0;
 }
 
-NabiIC*
-nabi_server_get_ic(NabiServer *server, CARD16 icid)
+NabiConnection*
+nabi_server_create_connection(NabiServer *server,
+			      CARD16 connect_id, const char* locale)
 {
-    if (icid > 0 && icid < server->ic_table_size)
-	return server->ic_table[icid];
-    else
+    NabiConnection* conn;
+    const char* encoding = NULL;
+
+    if (server == NULL)
 	return NULL;
-}
 
-void
-nabi_server_add_connect(NabiServer *server, NabiConnect *connect)
-{
-    NabiConnect *list;
-
-    if (server == NULL || connect == NULL)
-	return;
-
-    list = server->connect_list;
-    connect->next = list;
-    server->connect_list = connect;
-    server->n_connected++;
-}
-
-NabiConnect*
-nabi_server_get_connect_by_id(NabiServer *server, CARD16 connect_id)
-{
-    NabiConnect *connect;
-
-    connect = server->connect_list;
-    while (connect != NULL) {
-	if (connect->id == connect_id)
-	    return connect;
-	connect = connect->next;
+    if (locale != NULL) {
+	encoding = strchr(locale, '.');
+	if (encoding != NULL)
+	    encoding++; // skip '.'
     }
+
+    conn = nabi_connection_create(connect_id, encoding);
+    server->connections = g_slist_prepend(server->connections, conn);
+    return conn;
+}
+
+NabiConnection*
+nabi_server_get_connection(NabiServer *server, CARD16 connect_id)
+{
+    GSList* item;
+
+    item = server->connections;
+    while (item != NULL) {
+	NabiConnection* conn = (NabiConnection*)item->data;
+	if (conn != NULL && conn->id == connect_id)
+	    return conn;
+	item = g_slist_next(item);
+    }
+
     return NULL;
 }
 
 void
-nabi_server_remove_connect(NabiServer *server, NabiConnect *connect)
+nabi_server_destroy_connection(NabiServer *server, CARD16 connect_id)
 {
-    NabiConnect *prev;
-    NabiConnect *list;
+    NabiConnection* conn = nabi_server_get_connection(server, connect_id);
 
-    if (connect == NULL)
-	return;
-
-    prev = NULL;
-    list = server->connect_list;
-    while (list != NULL) {
-	if (list->id == connect->id) {
-	    if (prev == NULL)
-		server->connect_list = list->next;
-	    else
-		prev->next = list->next;
-	    list->next = NULL;
-	}
-	prev = list;
-	list = list->next;
-    }
-    server->n_connected--;
+    server->connections = g_slist_remove(server->connections, conn);
+    nabi_connection_destroy(conn);
 }
 
 Bool
@@ -673,34 +606,6 @@ nabi_server_is_locale_supported(NabiServer *server, const char *locale)
     return False;
 }
 
-Bool
-nabi_server_is_valid_char(NabiServer *server, wchar_t ch)
-{
-    int n;
-    gchar utf8[16];
-    gchar buf[16];
-    size_t ret;
-    gsize inbytesleft, outbytesleft;
-    gchar *inbuf;
-    gchar *outbuf;
-
-    if ((GIConv)server->converter == (GIConv)(-1))
-	return True;
-
-    n = hangul_wchar_to_utf8(ch, utf8, sizeof(utf8));
-    utf8[n] = '\0';
-
-    inbuf = utf8;
-    outbuf = buf;
-    inbytesleft = n;
-    outbytesleft = sizeof(buf);
-    ret = g_iconv(server->converter,
-	    	  &inbuf, &inbytesleft, &outbuf, &outbytesleft);
-    if ((GIConv)ret == (GIConv)(-1))
-	return False;
-    return True;
-}
-
 void
 nabi_server_on_keypress(NabiServer *server,
 			KeySym keyval,
@@ -722,348 +627,222 @@ nabi_server_on_keypress(NabiServer *server,
 	server->statistics.shift++;
 }
 
-static
-guint32 string_to_hex(char* p)
-{
-    guint32 ret = 0;
-    guint32 remain = 0;
-
-    if (*p == 'U')
-	p++;
-
-    while (*p != '\0') {
-	if (*p >= '0' && *p <= '9')
-	    remain = *p - '0';
-	else if (*p >= 'a' && *p <= 'f')
-	    remain = *p - 'a' + 10;
-	else if (*p >= 'A' && *p <= 'F')
-	    remain = *p - 'A' + 10;
-	else
-	    return 0;
-
-	ret = ret * 16 + remain;
-	p++;
-    }
-    return ret;
-}
-
-static gint
-candidate_item_compare(gconstpointer a, gconstpointer b)
-{
-    return ((NabiCandidateItem*)((GList*)a)->data)->ch
-	    - ((NabiCandidateItem*)((GList*)b)->data)->ch;
-}
-
-static guchar*
-skip_space(guchar* p)
+static gchar*
+skip_space(gchar* p)
 {
     while (g_ascii_isspace(*p))
 	p++;
     return p;
 }
 
-static gint
-compose_item_compare(gconstpointer a, gconstpointer b)
+static NabiKeyboardLayout*
+nabi_keyboard_layout_new(const char* name)
 {
-    return ((NabiComposeItem*)a)->key - ((NabiComposeItem*)b)->key;
+    NabiKeyboardLayout* layout = g_new(NabiKeyboardLayout, 1);
+    layout->name = g_strdup(name);
+    layout->table = NULL;
+    return layout;
 }
 
-static NabiComposeTable *
-create_compose_table_from_list(const char *name, GSList *list)
+static int
+nabi_keyboard_layout_cmp(const void* a, const void* b)
 {
-    int i;
-    NabiComposeTable *table = NULL;
-
-    table = g_new(NabiComposeTable, 1);
-    table->name = g_strdup(name);
-    table->items = NULL;
-    table->size = 0;
-
-    /* sort compose map */
-    list = g_slist_reverse(list);
-    list = g_slist_sort(list, compose_item_compare);
-
-    /* move data to map */
-    table->size = g_slist_length(list);
-    table->items = g_new(NabiComposeItem, table->size);
-    for (i = 0; i < table->size; i++, list = list->next) {
-	table->items[i] = *((NabiComposeItem*)list->data);
-	g_free(list->data);
-    }
-    return table;
+    const struct KeySymPair* pair1 = a;
+    const struct KeySymPair* pair2 = b;
+    return pair1->key - pair2->key;
 }
 
-Bool 
-nabi_server_load_keyboard_table(NabiServer *server, const char *filename)
+void
+nabi_keyboard_layout_append(NabiKeyboardLayout* layout,
+			    KeySym key, KeySym value)
 {
-    int i;
-    char *line, *p, *saved_position;
-    char buf[256];
-    FILE* file;
-    wchar_t key, value;
-    NabiKeyboardTable *table;
+    struct KeySymPair item = { key, value };
+    if (layout->table == NULL)
+	layout->table = g_array_new(FALSE, FALSE, sizeof(struct KeySymPair));
+    g_array_append_vals(layout->table, &item, 1);
+}
 
-    file = fopen(filename, "r");
-    if (file == NULL) {
-	fprintf(stderr, _("Nabi: Can't read keyboard map file\n"));
-	return False;
-    }
-
-    table = g_new(NabiKeyboardTable, 1);
-
-    /* init */
-    table->type = NABI_KEYBOARD_3SET;
-    table->filename = g_strdup(filename);
-    table->name = NULL;
-    table->output_mode = -1;
-    table->compose_table = NULL;
-
-    for (i = 0; i < sizeof(table->table) / sizeof(table->table[0]); i++)
-	table->table[i] = 0;
-
-    for (line = fgets(buf, sizeof(buf), file);
-	 line != NULL;
-	 line = fgets(buf, sizeof(buf), file)) {
-	p = strtok_r(line, " \t\n", &saved_position);
-	/* comment */
-	if (p == NULL || p[0] == '#')
-	    continue;
-
-	if (strcmp(p, "Name:") == 0) {
-	    p = strtok_r(NULL, "\n", &saved_position);
-	    if (p == NULL)
-		continue;
-	    table->name = g_strdup(p);
-	} else if (strcmp(p, "Compose:") == 0) {
-	    GList *list;
-	    p = strtok_r(NULL, " \n", &saved_position);
-	    if (p == NULL)
-		continue;
-
-	    for (list = server->compose_tables;
-		 list != NULL;
-		 list = g_list_next(list)) {
-		NabiComposeTable *compose_table = (NabiComposeTable*)list->data;
-		if (strcmp(compose_table->name, p) == 0) {
-		    table->compose_table = compose_table;
-		    break;
-		}
-	    }
-	} else if (strcmp(p, "Output:") == 0) {
-	    p = strtok_r(NULL, " \n", &saved_position);
-	    if (p == NULL)
-		continue;
-	    if (strcmp(p, "jamo") == 0) {
-		table->output_mode = NABI_OUTPUT_JAMO;
-	    } else if (strcmp(p, "syllable") == 0) {
-		table->output_mode = NABI_OUTPUT_SYLLABLE;
-	    } else if (strcmp(p, "manual") == 0) {
-		table->output_mode = NABI_OUTPUT_MANUAL;
-	    }
-	} else if (strcmp(p, "Type2") == 0) {
-	    table->type = NABI_KEYBOARD_2SET;
-	} else {
-	    key = string_to_hex(p);
-	    if (key == 0)
-		continue;
-
-	    p = strtok_r(NULL, " \t", &saved_position);
-	    if (p == NULL)
-		continue;
-	    value = string_to_hex(p);
-	    if (value == 0)
-		continue;
-
-	    if (key < XK_exclam || key > XK_asciitilde)
-		continue;
-
-	    table->table[key - XK_exclam] = value;
+KeySym
+nabi_keyboard_layout_get_key(NabiKeyboardLayout* layout, KeySym keysym)
+{
+    if (layout->table != NULL) {
+	struct KeySymPair key = { keysym, 0 };
+	struct KeySymPair* ret;
+	ret = bsearch(&key, layout->table->data, layout->table->len,
+		      sizeof(key), nabi_keyboard_layout_cmp);
+	if (ret) {
+	    return ret->value;
 	}
     }
-    fclose(file);
 
-    if (table->name == NULL)
-	table->name = g_path_get_basename(table->filename);
-
-    server->keyboard_tables = g_list_append(server->keyboard_tables,
-					    table);
-    return True;
+    return keysym;
 }
 
-Bool
-nabi_server_load_compose_table(NabiServer *server, const char *filename)
+static void
+nabi_keyboard_layout_free(gpointer data, gpointer user_data)
 {
-    char *line, *p, *saved_position, *name = NULL;
-    char buf[256];
-    FILE* file;
-    guint32 key1, key2;
-    wchar_t value;
-    NabiComposeTable *table = NULL;
-    NabiComposeItem *citem = NULL;
-    GSList *list = NULL;
-    Bool ret = False;
-
-    file = fopen(filename, "r");
-    if (file == NULL) {
-	fprintf(stderr, "Nabi: Can't open file: %s (%s)\n",
-		filename, strerror(errno));
-	return ret;
-    }
-
-    for (line = fgets(buf, sizeof(buf), file);
-	 line != NULL;
-	 line = fgets(buf, sizeof(buf), file)) {
-	p = strtok_r(line, " \t\n", &saved_position);
-	/* comment */
-	if (p == NULL || p[0] == '#')
-	    continue;
-
-	if (p[0] == '[') {
-	    char *end = strchr(p, ']');
-	    if (end == NULL)
-		continue;
-
-	    if (name != NULL && list != NULL) {
-		/* append new compose table */
-		table = create_compose_table_from_list(name, list);
-		server->compose_tables = g_list_append(server->compose_tables,
-						       table);
-		g_free(name);
-		name = NULL;
-		g_slist_free(list);
-		list = NULL;
-		ret = True;
-	    }
-	    *end = '\0';
-	    name = g_strdup(p + 1);
-	    continue;
-	} else {
-	    key1 = string_to_hex(p);
-	    if (key1 == 0)
-		continue;
-
-	    p = strtok_r(NULL, " \t", &saved_position);
-	    if (p == NULL)
-		continue;
-	    key2 = string_to_hex(p);
-	    if (key2 == 0)
-		continue;
-
-	    p = strtok_r(NULL, " \t", &saved_position);
-	    if (p == NULL)
-		continue;
-	    value = string_to_hex(p);
-	    if (value == 0)
-		continue;
-
-	    citem = g_new(NabiComposeItem, 1);
-	    citem->key = key1 << 16 | key2;
-	    citem->code = value;
-
-	    list = g_slist_prepend(list, citem);
-	}
-    }
-    fclose(file);
-
-    if (name != NULL && list != NULL) {
-	table = create_compose_table_from_list(name, list);
-	server->compose_tables = g_list_append(server->compose_tables,
-					       table);
-	g_free(name);
-	name = NULL;
-	g_slist_free(list);
-	list = NULL;
-	ret = True;
-    }
-
-    return ret;
+    NabiKeyboardLayout *layout = data;
+    g_free(layout->name);
+    if (layout->table != NULL)
+	g_array_free(layout->table, TRUE);
+    g_free(layout);
 }
 
-Bool
-nabi_server_load_candidate_table(NabiServer *server,
-				 const char *filename)
+static void
+nabi_server_delete_layouts(NabiServer* server)
+{
+    if (server->layouts != NULL) {
+	g_list_foreach(server->layouts, nabi_keyboard_layout_free, NULL);
+	g_list_free(server->layouts);
+	server->layouts = NULL;
+    }
+}
+
+void
+nabi_server_load_keyboard_layout(NabiServer *server, const char *filename)
 {
     FILE *file;
-    gint n, i, j;
-    unsigned char *p;
-    unsigned char buf[256];
-    gunichar ch;
-    gunichar key;
-    NabiCandidateItem *item;
-    GList *table = NULL;
-    GList *items = NULL;
-    GList *pitem = NULL;
+    char *p, *line;
+    char *saved_position = NULL;
+    char buf[256];
+    GList *list = NULL;
+    NabiKeyboardLayout *layout = NULL;
 
     file = fopen(filename, "r");
     if (file == NULL) {
-	fprintf(stderr, "Nabi: Failed to open candidate file: %s\n", filename);
-	return False;
+	fprintf(stderr, "Nabi: Failed to open keyboard layout file: %s\n", filename);
+	return;
     }
 
-    for (p = fgets(buf, sizeof(buf), file);
-	 p != NULL;
-	 p = fgets(buf, sizeof(buf), file)) {
+    layout = nabi_keyboard_layout_new("none");
+    list = g_list_append(list, layout);
+    layout = NULL;
 
-	buf[sizeof(buf) - 1] = '\0';
-	p = skip_space(p);
+    for (line = fgets(buf, sizeof(buf), file);
+	 line != NULL;
+	 line = fgets(buf, sizeof(buf), file)) {
+	p = skip_space(buf);
 
 	/* skip comments */
 	if (*p == '\0' || *p == ';' || *p == '#')
 	    continue;
 
-	if (*p == '[') {
-	    p++;
-	    if (items != NULL) {
-		items = g_list_reverse(items);
-		table = g_list_prepend(table, (gpointer)items);
-	    }
-	    key = g_utf8_get_char_validated(p, sizeof(buf));
-	    if ((gunichar)key != (gunichar)-2 &&
-		(gunichar)key != (gunichar)-1) {
-		item = nabi_candidate_item_new(key, NULL);
-		items = NULL;
-		items = g_list_prepend(items, item);
-	    }
-	} else {
-	    ch = g_utf8_get_char(p);
-	    p = strchr(p, '=');
+	if (p[0] == '[') {
+	    p = strtok_r(p + 1, "]", &saved_position);
 	    if (p != NULL) {
-		p = g_strchomp(p + 1);
-		item = nabi_candidate_item_new(ch, p);
-		items = g_list_prepend(items, (gpointer)item);
+		if (layout != NULL)
+		    list = g_list_append(list, layout);
+
+		layout = nabi_keyboard_layout_new(p);
 	    }
+	} else if (layout != NULL) {
+	    KeySym key, value;
+
+	    p = strtok_r(p, " \t", &saved_position);
+	    if (p == NULL)
+		continue;
+
+	    key = strtol(p, NULL, 16);
+	    if (key == 0)
+		continue;
+
+	    p = strtok_r(NULL, "\r\n\t ", &saved_position);
+	    if (p == NULL)
+		continue;
+
+	    value = strtol(p, NULL, 16);
+	    if (value == 0)
+		continue;
+
+	    nabi_keyboard_layout_append(layout, key, value);
 	}
     }
-    if (items != NULL) {
-	items = g_list_reverse(items);
-	table = g_list_prepend(table, (gpointer)items);
-	items = NULL;
-    }
+
+    if (layout != NULL)
+	list = g_list_append(list, layout);
 
     fclose(file);
 
-    table = g_list_reverse(table);
-    table = g_list_sort(table, candidate_item_compare);
+    nabi_server_delete_layouts(server);
+    server->layouts = list;
+}
 
-    server->candidate_table_size = g_list_length(table);
-    server->candidate_table = g_new(NabiCandidateItem**,
-				    server->candidate_table_size);
-    items = table;
-    for (i = 0; i < server->candidate_table_size; i++) {
-	pitem = items->data;
-	n = g_list_length(pitem) + 1;
-	server->candidate_table[i] = g_new(NabiCandidateItem*, n);
-	for (j = 0;  pitem != NULL; j++) {
-	    server->candidate_table[i][j] = pitem->data;
-	    pitem = g_list_next(pitem);
-	}
-	server->candidate_table[i][j] = NULL;
-	g_list_free((GList*)items->data);
-	items = g_list_next(items);
+void
+nabi_server_set_keyboard_layout(NabiServer *server, const char* name)
+{
+    GList* list;
+
+    if (strcmp(name, "none") == 0) {
+	server->layout = NULL;
+	return;
     }
-    g_list_free(table);
 
-    return TRUE;
+    list = server->layouts;
+    while (list != NULL) {
+	NabiKeyboardLayout* layout = list->data;
+	if (strcmp(layout->name, name) == 0) {
+	    server->layout = layout;
+	}
+	list = g_list_next(list);
+    }
+}
+
+const NabiHangulKeyboard*
+nabi_server_get_hangul_keyboard_list(NabiServer* server)
+{
+    return server->hangul_keyboard_list;
+}
+
+KeySym
+nabi_server_normalize_keysym(NabiServer *server,
+			     KeySym keysym, unsigned int state)
+{
+    KeySym upper, lower;
+
+    /* unicode keysym */
+    if ((keysym & 0xff000000) == 0x01000000)
+	keysym &= 0x00ffffff;
+
+    /* european mapping */
+    if (server->layout != NULL) {
+	keysym = nabi_keyboard_layout_get_key(server->layout, keysym);
+    }
+
+    upper = keysym;
+    lower = keysym;
+    XConvertCase(keysym, &lower, &upper);
+
+    if (state & ShiftMask)
+	keysym = upper;
+    else
+	keysym = lower;
+
+    return keysym;
+}
+
+void
+nabi_server_set_dynamic_event_flow(NabiServer* server, Bool flag)
+{
+    server->dynamic_event_flow = flag;
+}
+
+void
+nabi_server_set_xim_name(NabiServer* server, const char* name)
+{
+    g_free(server->name);
+    server->name = g_strdup(name);
+}
+
+void
+nabi_server_set_commit_by_word(NabiServer* server, Bool flag)
+{
+    server->commit_by_word = flag;
+}
+
+void
+nabi_server_set_auto_reorder(NabiServer* server, Bool flag)
+{
+    server->auto_reorder = flag;
 }
 
 void
@@ -1089,7 +868,7 @@ nabi_server_write_log(NabiServer *server)
 	fprintf(file, "total: %d\n", server->statistics.total);
 	fprintf(file, "space: %d\n", server->statistics.space);
 	fprintf(file, "backspace: %d\n", server->statistics.backspace);
-	fprintf(file, "keyboard: %s\n", server->keyboard_table->name);
+	fprintf(file, "keyboard: %s\n", server->hangul_keyboard);
 
 	fprintf(file, "cho:  "
 		      "\341\204\200: %-3d "
