@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004 Choe Hwanjin
+ * Copyright (C) 2003,2004,2007 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -28,36 +28,26 @@
 #include <glib.h>
 #include <gtk/gtk.h>
 
+#include <hangul.h>
+
 #include "../IMdkit/IMdkit.h"
 #include "../IMdkit/Xi18n.h"
 
 #include "ic.h"
-#include "candidate.h"
 
-typedef struct _NabiKeyboardTable NabiKeyboardTable;
-typedef struct _NabiComposeItem NabiComposeItem;
-typedef struct _NabiComposeTable NabiComposeTable;
+typedef struct _NabiHangulKeyboard NabiHangulKeyboard;
+typedef struct _NabiKeyboardLayout NabiKeyboardLayout;
 typedef struct _NabiServer NabiServer;
 
 #define KEYBOARD_TABLE_SIZE 94
-struct _NabiKeyboardTable {
-    gchar*            filename;
-    gchar*            name;
-    gint              type;
-    int		      output_mode;
-    NabiComposeTable* compose_table;
-    wchar_t           table[KEYBOARD_TABLE_SIZE];
+struct _NabiHangulKeyboard {
+    const gchar* id;
+    const gchar* name;
 };
 
-struct _NabiComposeItem {
-    uint32_t key;
-    wchar_t code;
-};
-
-struct _NabiComposeTable {
-    gchar           *name;
-    NabiComposeItem *items;
-    gint            size;
+struct _NabiKeyboardLayout {
+    char*  name;
+    GArray* table;
 };
 
 enum {
@@ -113,34 +103,32 @@ struct _NabiServer {
     GdkGC*		    gc;
 
     /* xim connect list */
-    int			    n_connected;
-    NabiConnect*            connect_list;
+    GSList*                 connections;
 
     /* Input Context list */
+    CARD16                  last_icid;
+    GSList*                 freed_icid;
     NabiIC**                ic_table;
     int                     ic_table_size;
-    NabiIC*                 ic_freed;
+
+    /* keyboard translate */
+    GList*                  layouts;
+    NabiKeyboardLayout*     layout;
 
     /* hangul automata */
-    GList*		    keyboard_tables;
-    GList*		    compose_tables;
-    NabiKeyboardTable*	    keyboard_table;
-    NabiComposeTable*	    compose_table;
-    Bool                    (*automata)(NabiIC*,
-                                        KeySym,
-                                        unsigned int state);
-    Bool                    dvorak;
-    NabiOutputMode          output_mode;
-    int			    candidate_table_size;
-    NabiCandidateItem***    candidate_table;
+    char*                   hangul_keyboard;
+    const NabiHangulKeyboard* hangul_keyboard_list;
 
-    /* hangul converter */
-    Bool                    check_charset;
-    GIConv                  converter;
+    NabiOutputMode          output_mode;
+
+    /* hanja */
+    HanjaTable*             hanja_table;
 
     /* options */
     Bool                    dynamic_event_flow;
     Bool                    global_input_mode;
+    Bool                    commit_by_word;
+    Bool                    auto_reorder;
     Bool                    show_status;
     NabiInputMode           input_mode;
     GdkColor                preedit_fg;
@@ -172,32 +160,43 @@ Bool        nabi_server_is_candidate_key(NabiServer*  server,
 void        nabi_server_set_trigger_keys(NabiServer *server,
 					 char **keys);
 void        nabi_server_set_candidate_keys(NabiServer *server, char **keys);
-void        nabi_server_set_dvorak      (NabiServer *server,
-                                         Bool flag);
-void        nabi_server_set_keyboard_table(NabiServer *server,
-					 const char *name);
-void        nabi_server_set_compose_table(NabiServer *server,
-					 const char *name);
+
+void        nabi_server_load_keyboard_layout(NabiServer *server,
+					     const char *filename);
+void        nabi_server_set_keyboard_layout(NabiServer *server,
+					    const char* name);
+void        nabi_server_set_hangul_keyboard(NabiServer *server,
+					 const char *id);
+
 void        nabi_server_set_mode_info_cb(NabiServer *server,
 					 NabiModeInfoCallback func);
 void        nabi_server_set_output_mode (NabiServer *server,
 					 NabiOutputMode mode);
 void	    nabi_server_set_candidate_font(NabiServer *server,
 					   const gchar *font_name);
+void        nabi_server_set_dynamic_event_flow(NabiServer* server, Bool flag);
+void        nabi_server_set_xim_name(NabiServer* server, const char* name);
+void        nabi_server_set_commit_by_word(NabiServer* server, Bool flag);
+void        nabi_server_set_auto_reorder(NabiServer* server, Bool flag);
 
-void        nabi_server_ic_table_expand (NabiServer* server);
-NabiIC*     nabi_server_get_ic          (NabiServer *server,
-                                         CARD16 icid);
+NabiIC*     nabi_server_alloc_ic        (NabiServer* server);
+void        nabi_server_dealloc_ic      (NabiServer* server, NabiIC* ic);
+NabiIC*     nabi_server_get_ic          (NabiServer *server, CARD16 icid);
+gboolean    nabi_server_is_valid_ic     (NabiServer* server, NabiIC* ic);
+
+NabiConnection* nabi_server_create_connection (NabiServer *server,
+					       CARD16 connect_id,
+					       const char* locale);
+NabiConnection* nabi_server_get_connection    (NabiServer *server,
+					       CARD16 connect_id);
+void            nabi_server_destroy_connection(NabiServer *server,
+					       CARD16 connect_id);
 
-void        nabi_server_add_connect     (NabiServer *server,
-	                                 NabiConnect *connect);
-void        nabi_server_remove_connect  (NabiServer *server,
-                                         NabiConnect *connect);
-NabiConnect* nabi_server_get_connect_by_id(NabiServer *server,
-                                           CARD16 connect_id);
 Bool        nabi_server_is_locale_supported(NabiServer *server,
 					    const char *locale);
-Bool        nabi_server_is_valid_char   (NabiServer *server, wchar_t ch);
+Bool        nabi_server_is_valid_str    (NabiServer *server, const char* str);
+KeySym      nabi_server_normalize_keysym(NabiServer *server,
+					 KeySym keysym, unsigned int state);
 void        nabi_server_on_keypress     (NabiServer *server,
 					 KeySym keyval,
 					 unsigned int state,
@@ -208,8 +207,9 @@ Bool	    nabi_server_load_keyboard_table(NabiServer *server,
 					    const char *filename);
 Bool	    nabi_server_load_compose_table(NabiServer *server,
 					   const char *filename);
-Bool        nabi_server_load_candidate_table(NabiServer *server,
-				             const char *filename);
+
+const NabiHangulKeyboard*
+nabi_server_get_hangul_keyboard_list(NabiServer* server);
 
 #endif  /* __SERVER_H_ */
 
