@@ -1,5 +1,5 @@
 /* Nabi - X Input Method server for hangul
- * Copyright (C) 2003,2004 Choe Hwanjin
+ * Copyright (C) 2003,2004,2007 Choe Hwanjin
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -34,10 +34,12 @@
 
 #include "eggtrayicon.h"
 
+#include "debug.h"
 #include "gettext.h"
+#include "nabi.h"
 #include "ic.h"
 #include "server.h"
-#include "nabi.h"
+#include "conf.h"
 
 #include "default-icons.h"
 
@@ -73,294 +75,6 @@ static GtkWidget *none_image = NULL;
 static GtkWidget *hangul_image = NULL;
 static GtkWidget *english_image = NULL;
 
-
-enum {
-    CONF_TYPE_BOOL,
-    CONF_TYPE_INT,
-    CONF_TYPE_STR
-};
-
-struct config_item {
-    gchar* key;
-    gint   type;
-    guint  offset;
-};
-
-#define BASE           ((void*)((NabiApplication*)NULL))
-#define MEMBER(member) ((void*)(&(((NabiApplication*)NULL)->member)))
-#define OFFSET(member) ((guint)(MEMBER(member) - BASE))
-
-#define BOOL_BY_OFFSET(offset) (gboolean*)((void*)(nabi) + offset)
-#define INT_BY_OFFSET(offset)      (gint*)((void*)(nabi) + offset)
-#define CHAR_BY_OFFSET(offset)   (gchar**)((void*)(nabi) + offset)
-
-const static struct config_item config_items[] = {
-    { "xim_name",           CONF_TYPE_STR,  OFFSET(xim_name)                 },
-    { "x",                  CONF_TYPE_INT,  OFFSET(x)                        },
-    { "y",                  CONF_TYPE_INT,  OFFSET(y)                        },
-    { "theme",              CONF_TYPE_STR,  OFFSET(theme)                    },
-    { "keyboard_table_name",CONF_TYPE_STR,  OFFSET(keyboard_table_name)      },
-    { "keyboard_table_dir", CONF_TYPE_STR,  OFFSET(keyboard_table_dir)       },
-    { "compose_table_name", CONF_TYPE_STR,  OFFSET(compose_table_name)       },
-    { "compose_table_dir",  CONF_TYPE_STR,  OFFSET(compose_table_dir)        },
-    { "candidate_table",    CONF_TYPE_STR,  OFFSET(candidate_table_filename) },
-    { "triggerkeys",        CONF_TYPE_STR,  OFFSET(trigger_keys)             },
-    { "candidatekeys",      CONF_TYPE_STR,  OFFSET(candidate_keys)           },
-    { "dvorak",             CONF_TYPE_BOOL, OFFSET(dvorak)                   },
-    { "output_mode",        CONF_TYPE_STR,  OFFSET(output_mode)              },
-    { "preedit_foreground", CONF_TYPE_STR,  OFFSET(preedit_fg)               },
-    { "preedit_background", CONF_TYPE_STR,  OFFSET(preedit_bg)               },
-    { "candidate_font",	    CONF_TYPE_STR,  OFFSET(candidate_font)           },
-    { NULL,                 0,              0                                }
-};
-
-static void
-set_value_bool(guint offset, gchar* value)
-{
-    gboolean *member = BOOL_BY_OFFSET(offset);
-
-    if (g_ascii_strcasecmp(value, "true") == 0) {
-	*member = TRUE;
-    } else {
-	*member = FALSE;
-    }
-}
-
-static void
-set_value_int(guint offset, gchar* value)
-{
-    int *member = INT_BY_OFFSET(offset);
-
-    *member = strtol(value, NULL, 10);
-}
-
-static void
-set_value_str(guint offset, gchar* value)
-{
-    char **member = CHAR_BY_OFFSET(offset);
-
-    g_free(*member);
-    if (value == NULL)
-	*member = g_strdup("");
-    else
-	*member = g_strdup(value);
-}
-
-static void
-write_value_bool(FILE* file, gchar* key, guint offset)
-{
-    gboolean *member = BOOL_BY_OFFSET(offset);
-
-    if (*member)
-	fprintf(file, "%s=%s\n", key, "true");
-    else
-	fprintf(file, "%s=%s\n", key, "false");
-}
-
-static void
-write_value_int(FILE* file, gchar* key, guint offset)
-{
-    int *member = INT_BY_OFFSET(offset);
-
-    fprintf(file, "%s=%d\n", key, *member);
-}
-
-static void
-write_value_str(FILE* file, gchar* key, guint offset)
-{
-    char **member = CHAR_BY_OFFSET(offset);
-
-    if (*member != NULL)
-	fprintf(file, "%s=%s\n", key, *member);
-}
-
-static void
-load_config_item(gchar* key, gchar* value)
-{
-    gint i;
-
-    for (i = 0; config_items[i].key != NULL; i++) {
-	if (strcmp(key, config_items[i].key) == 0) {
-	    switch (config_items[i].type) {
-	    case CONF_TYPE_BOOL:
-		if (value != NULL)
-		    set_value_bool(config_items[i].offset, value);
-		break;
-	    case CONF_TYPE_INT:
-		if (value != NULL)
-		    set_value_int(config_items[i].offset, value);
-		break;
-	    case CONF_TYPE_STR:
-		set_value_str(config_items[i].offset, value);
-		break;
-	    default:
-		break;
-	    }
-	}
-    }
-}
-
-static void
-load_config_file(void)
-{
-    gchar *line, *saved_position;
-    gchar *key, *value;
-    const gchar* homedir;
-    gchar* config_filename;
-    gchar buf[256];
-    FILE *file;
-
-    /* set default values */
-    nabi->xim_name = g_strdup(PACKAGE);
-    nabi->theme = g_strdup("SimplyRed");
-    nabi->icon_size = 24;
-    nabi->keyboard_table_name = g_strdup(DEFAULT_KEYBOARD);
-    nabi->keyboard_table_dir = g_build_filename(NABI_DATA_DIR,
-						"keyboard",
-						NULL);
-    nabi->compose_table_name = g_strdup("default");
-    nabi->compose_table_dir = g_build_filename(NABI_DATA_DIR,
-					       "compose",
-					       NULL);
-    nabi->candidate_table_filename = g_build_filename(NABI_DATA_DIR,
-						      "candidate",
-						      "nabi.txt",
-						      NULL);
-    nabi->trigger_keys = g_strdup("Hangul,Shift+space");
-    nabi->candidate_keys = g_strdup("Hangul_Hanja,F9");
-
-    nabi->candidate_font = g_strdup("Sans 14");
-    nabi->output_mode = g_strdup("syllable");
-    nabi->preedit_fg = g_strdup("#FFFFFF");
-    nabi->preedit_bg = g_strdup("#000000");
-
-    /* load conf file */
-    homedir = g_get_home_dir();
-    config_filename = g_build_filename(homedir, ".nabi", "config", NULL);
-    file = fopen(config_filename, "r");
-    if (file == NULL) {
-	fprintf(stderr, _("Nabi: Can't load config file\n"));
-	return;
-    }
-
-    for (line = fgets(buf, sizeof(buf), file);
-	 line != NULL;
-	 line = fgets(buf, sizeof(buf), file)) {
-	key = strtok_r(line, " =\t\n", &saved_position);
-	value = strtok_r(NULL, "\r\n", &saved_position);
-	if (key == NULL)
-	    continue;
-	load_config_item(key, value);
-    }
-    fclose(file);
-    g_free(config_filename);
-}
-
-void
-nabi_save_config_file(void)
-{
-    gint i;
-    const gchar* homedir;
-    gchar* config_dir;
-    gchar* config_filename;
-    FILE *file;
-
-    homedir = g_get_home_dir();
-    config_dir = g_build_filename(homedir, ".nabi", NULL);
-
-    /* chech for nabi conf dir */
-    if (!g_file_test(config_dir, G_FILE_TEST_EXISTS)) {
-	int ret;
-	/* we make conf dir */
-	ret = mkdir(config_dir, S_IRUSR | S_IWUSR | S_IXUSR);
-	if (ret != 0) {
-	    perror("nabi");
-	}
-    }
-
-    config_filename = g_build_filename(config_dir, "config", NULL);
-    file = fopen(config_filename, "w");
-    if (file == NULL) {
-	fprintf(stderr, _("Nabi: Can't write config file\n"));
-	return;
-    }
-
-    for (i = 0; config_items[i].key != NULL; i++) {
-	switch (config_items[i].type) {
-	case CONF_TYPE_BOOL:
-	    write_value_bool(file, config_items[i].key,
-			   config_items[i].offset);
-	    break;
-	case CONF_TYPE_INT:
-	    write_value_int(file, config_items[i].key,
-			  config_items[i].offset);
-	    break;
-	case CONF_TYPE_STR:
-	    write_value_str(file, config_items[i].key,
-			  config_items[i].offset);
-	    break;
-	default:
-	    break;
-	}
-    }
-    fclose(file);
-
-    g_free(config_dir);
-    g_free(config_filename);
-}
-
-static void
-load_compose_table(void)
-{
-    Bool ret;
-    const char *path;
-    gchar *filename;
-    DIR *dir;
-    struct dirent *dent;
-    gboolean flag = FALSE;
-
-    path = nabi->compose_table_dir;
-    dir = opendir(path);
-    if (dir == NULL) {
-	fprintf(stderr,
-		_("Nabi: Can't open compose table dir: %s (%s)\n"),
-		path, strerror(errno));
-	return;
-    }
-
-    for (dent = readdir(dir); dent != NULL; dent = readdir(dir)) {
-	if (dent->d_name[0] == '.')
-	    continue;
-
-	filename = g_build_filename(path, dent->d_name, NULL);
-	ret = nabi_server_load_compose_table(nabi_server, filename);
-	if (ret) {
-	    flag = TRUE;
-	} else {
-	    fprintf(stderr, _("Nabi: Can't load compose table: %s\n"),filename);
-	}
-	g_free(filename);
-    }
-    closedir(dir);
-
-    if (!flag) {
-	/* we have no valid compose table */
-	fprintf(stderr, _("Nabi: Can't load any compose table\n"));
-    }
-    nabi_server_set_compose_table(nabi_server, nabi->compose_table_name);
-}
-
-static void
-load_candidate_table(void)
-{
-    Bool ret;
-    ret = nabi_server_load_candidate_table(nabi_server,
-					   nabi->candidate_table_filename);
-    if (!ret) {
-	fprintf(stderr, _("Nabi: Can't load candidate table\n"));
-    }
-}
-
 static void
 load_colors(void)
 {
@@ -371,7 +85,7 @@ load_colors(void)
     colormap = gdk_colormap_get_system();
 
     /* preedit foreground */
-    gdk_color_parse(nabi->preedit_fg, &color);
+    gdk_color_parse(nabi->config->preedit_fg, &color);
     ret = gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);
     if (ret)
 	nabi_server->preedit_fg = color;
@@ -385,7 +99,7 @@ load_colors(void)
     }
 
     /* preedit background */
-    gdk_color_parse(nabi->preedit_bg, &color);
+    gdk_color_parse(nabi->config->preedit_bg, &color);
     ret = gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);
     if (ret)
 	nabi_server->preedit_bg = color;
@@ -402,84 +116,41 @@ load_colors(void)
 static void
 set_up_keyboard(void)
 {
-    Bool ret;
-    const char *path;
-    gchar *filename;
-    DIR *dir;
-    struct dirent *dent;
-    gboolean flag = FALSE;
-
-    /* dvorak set */
-    nabi_server_set_dvorak(nabi_server, nabi->dvorak);
-
-    /* load keyboard tables from keyboard table dir */
-    path = nabi->keyboard_table_dir;
-    dir = opendir(path);
-    if (dir == NULL) {
-	fprintf(stderr, _("Nabi: Can't open keyboard table dir: %s\n"), path);
-	return;
-    }
-
-    for (dent = readdir(dir); dent != NULL; dent = readdir(dir)) {
-	if (dent->d_name[0] == '.')
-	    continue;
-
-	filename = g_build_filename(path, dent->d_name, NULL);
-	ret = nabi_server_load_keyboard_table(nabi_server, filename);
-	if (ret) {
-	    flag = TRUE;
-	} else {
-	    fprintf(stderr,
-		    _("Nabi: Can't load keyboard table: %s\n"), filename);
-	}
+    /* keyboard layout */
+    if (g_path_is_absolute(nabi->config->keyboard_layouts_file)) {
+	nabi_server_load_keyboard_layout(nabi_server,
+					 nabi->config->keyboard_layouts_file);
+    } else {
+	char* filename = g_build_filename(NABI_DATA_DIR,
+				  nabi->config->keyboard_layouts_file, NULL);
+	nabi_server_load_keyboard_layout(nabi_server,
+					 filename);
 	g_free(filename);
     }
-    closedir(dir);
-
-    if (!flag) {
-	/* we have no valid keyboard table */
-	fprintf(stderr, _("Nabi: Can't load any keyboard table\n"));
-    }
+    nabi_server_set_keyboard_layout(nabi_server, nabi->config->latin_keyboard);
 
     /* set keyboard */
-    nabi_server_set_keyboard_table(nabi_server, nabi->keyboard_table_name);
+    nabi_server_set_hangul_keyboard(nabi_server, nabi->config->hangul_keyboard);
 }
 
 void
 nabi_app_new(void)
 {
-    nabi = g_malloc(sizeof(NabiApplication));
+    nabi = g_new(NabiApplication, 1);
 
     nabi->xim_name = NULL;
-    nabi->optional_xim_name = NULL;
-    nabi->x = 0;
-    nabi->y = 0;
     nabi->main_window = NULL;
     nabi->status_only = FALSE;
     nabi->session_id = NULL;
-
-    nabi->theme = NULL;
-    nabi->keyboard_table_name = NULL;
-    nabi->keyboard_table_dir = NULL;
-    nabi->compose_table_name = NULL;
-    nabi->compose_table_dir = NULL;
-    nabi->candidate_table_filename = NULL;
-
-    nabi->trigger_keys = NULL;
-    nabi->candidate_keys = NULL;
-
-    nabi->dvorak = FALSE;
-    nabi->output_mode = NULL;
-
-    nabi->preedit_fg = NULL;
-    nabi->preedit_bg = NULL;
-    nabi->candidate_font = NULL;
+    nabi->icon_size = 0;
 
     nabi->root_window = NULL;
 
     nabi->mode_info_atom = 0;
     nabi->mode_info_type = 0;
     nabi->mode_info_xatom = 0;
+
+    nabi->config = NULL;
 }
 
 void
@@ -526,7 +197,14 @@ nabi_app_init(int *argc, char ***argv)
 		}
 		(*argv)[i] = NULL;
 
-		nabi->optional_xim_name = g_strdup(xim_name);
+		nabi->xim_name = g_strdup(xim_name);
+	    } else if (strcmp("-d", (*argv)[i]) == 0) {
+		gchar* log_level = "0";
+		(*argv)[i] = NULL;
+		i++;
+		log_level = (*argv)[i];
+
+		nabi_log_set_level(strtol(log_level, NULL, 10));
 	    }
 	    i++;
 	}
@@ -545,7 +223,8 @@ nabi_app_init(int *argc, char ***argv)
 	}
     }
 
-    load_config_file();
+    nabi->config = nabi_config_new();
+    nabi_config_load(nabi->config);
 
     /* set atoms for hangul status */
     nabi->mode_info_atom = gdk_atom_intern("_HANGUL_INPUT_MODE", TRUE);
@@ -553,8 +232,14 @@ nabi_app_init(int *argc, char ***argv)
     nabi->mode_info_xatom = gdk_x11_atom_to_xatom(nabi->mode_info_atom);
 
     /* default icon */
+    nabi->icon_size = 24;
     icon_filename = g_build_filename(NABI_DATA_DIR, "nabi.svg", NULL);
     default_icon = gdk_pixbuf_new_from_file(icon_filename, NULL);
+    if (default_icon == NULL) {
+	g_free(icon_filename);
+	icon_filename = g_build_filename(NABI_DATA_DIR, "nabi.png", NULL);
+	default_icon = gdk_pixbuf_new_from_file(icon_filename, NULL);
+    }
     g_free(icon_filename);
 }
 
@@ -564,10 +249,10 @@ set_up_output_mode(void)
     NabiOutputMode mode;
 
     mode = NABI_OUTPUT_SYLLABLE;
-    if (nabi->output_mode != NULL) {
-	if (g_ascii_strcasecmp(nabi->output_mode, "jamo") == 0) {
+    if (nabi->config->output_mode != NULL) {
+	if (g_ascii_strcasecmp(nabi->config->output_mode, "jamo") == 0) {
 	    mode = NABI_OUTPUT_JAMO;
-	} else if (g_ascii_strcasecmp(nabi->output_mode, "manual") == 0) {
+	} else if (g_ascii_strcasecmp(nabi->config->output_mode, "manual") == 0) {
 	    mode = NABI_OUTPUT_MANUAL;
 	}
     }
@@ -605,18 +290,20 @@ nabi_app_setup_server(void)
 	gtk_widget_destroy(message);
     }
 
-    load_compose_table();
-    load_candidate_table();
     set_up_keyboard();
     load_colors();
     set_up_output_mode();
-    keys = g_strsplit(nabi->trigger_keys, ",", 0);
+    keys = g_strsplit(nabi->config->trigger_keys, ",", 0);
     nabi_server_set_trigger_keys(nabi_server, keys);
     g_strfreev(keys);
-    keys = g_strsplit(nabi->candidate_keys, ",", 0);
+    keys = g_strsplit(nabi->config->candidate_keys, ",", 0);
     nabi_server_set_candidate_keys(nabi_server, keys);
     g_strfreev(keys);
-    nabi_server_set_candidate_font(nabi_server, nabi->candidate_font);
+    nabi_server_set_candidate_font(nabi_server, nabi->config->candidate_font);
+    nabi_server_set_dynamic_event_flow(nabi_server,
+				       nabi->config->use_dynamic_event_flow);
+    nabi_server_set_commit_by_word(nabi_server, nabi->config->commit_by_word);
+    nabi_server_set_auto_reorder(nabi_server, nabi->config->auto_reorder);
 }
 
 void
@@ -632,7 +319,7 @@ nabi_app_quit(void)
 
     if (nabi != NULL && nabi->main_window != NULL) {
 	gtk_window_get_position(GTK_WINDOW(nabi->main_window),
-				&nabi->x, &nabi->y);
+				&nabi->config->x, &nabi->config->y);
 	gtk_widget_destroy(nabi->main_window);
 	nabi->main_window = NULL;
     }
@@ -650,30 +337,16 @@ nabi_app_free(void)
     if (nabi == NULL)
 	return;
 
-    nabi_save_config_file();
+    nabi_app_save_config();
+    nabi_config_delete(nabi->config);
+    nabi->config = NULL;
 
     if (nabi->main_window != NULL) {
 	gtk_widget_destroy(nabi->main_window);
 	nabi->main_window = NULL;
     }
 
-    g_free(nabi->theme);
-
-    /* keyboard map */
-    g_free(nabi->keyboard_table_name);
-    g_free(nabi->keyboard_table_dir);
-
-    /* compose_table */
-    g_free(nabi->compose_table_name);
-    g_free(nabi->compose_table_dir);
-
-    g_free(nabi->output_mode);
-
-    g_free(nabi->preedit_fg);
-    g_free(nabi->preedit_bg);
-
     g_free(nabi->xim_name);
-    g_free(nabi->optional_xim_name);
 
     g_free(nabi);
     nabi = NULL;
@@ -686,7 +359,7 @@ on_tray_icon_embedded(GtkWidget *widget, gpointer data)
 	nabi->main_window != NULL &&
 	GTK_WIDGET_VISIBLE(nabi->main_window)) {
 	gtk_window_get_position(GTK_WINDOW(nabi->main_window),
-				&nabi->x, &nabi->y);
+				&nabi->config->x, &nabi->config->y);
 	gtk_widget_hide(GTK_WIDGET(nabi->main_window));
     }
 }
@@ -713,7 +386,8 @@ on_tray_icon_destroyed(GtkWidget *widget, gpointer data)
     if (nabi != NULL &&
 	nabi->main_window != NULL &&
 	!GTK_WIDGET_VISIBLE(nabi->main_window)) {
-	gtk_window_move(GTK_WINDOW(nabi->main_window), nabi->x, nabi->y);
+	gtk_window_move(GTK_WINDOW(nabi->main_window),
+			nabi->config->x, nabi->config->y);
 	gtk_widget_show(GTK_WIDGET(nabi->main_window));
     }
 }
@@ -782,10 +456,11 @@ on_tray_icon_button_press(GtkWidget *widget,
     case 2:
 	if (GTK_WIDGET_VISIBLE(nabi->main_window)) {
 	    gtk_window_get_position(GTK_WINDOW(nabi->main_window),
-			    &nabi->x, &nabi->y);
+			    &nabi->config->x, &nabi->config->y);
 	    gtk_widget_hide(GTK_WIDGET(nabi->main_window));
 	} else {
-	    gtk_window_move(GTK_WINDOW(nabi->main_window), nabi->x, nabi->y);
+	    gtk_window_move(GTK_WINDOW(nabi->main_window),
+			    nabi->config->x, nabi->config->y);
 	    gtk_widget_show(GTK_WIDGET(nabi->main_window));
 	}
 	return TRUE;
@@ -823,186 +498,81 @@ on_tray_icon_size_allocate (GtkWidget *widget,
     }
 }
 
-static void get_statistic_string(char *buf, size_t bufsize)
+static void get_statistic_string(GString *str)
 {
     if (nabi_server == NULL) {
-	snprintf(buf, bufsize, _("XIM Server is not running"));
+	g_string_assign(str, _("XIM Server is not running"));
     } else {
-	char general[256];
-	char choseong[256];
-	char jungseong[256];
-	char jongseong[256];
+	int i;
+	int sum;
 
-	snprintf(general, sizeof(general), 
+	g_string_append_printf(str, 
 		 "%s: %3d\n"
 		 "%s: %3d\n"
 		 "%s: %3d\n"
 		 "%s: %3d\n"
 		 "\n",
-		 _("Total"),
-		 nabi_server->statistics.total,
-		 _("Space"),
-		 nabi_server->statistics.space,
-		 _("BackSpace"),
-		 nabi_server->statistics.backspace,
-		 _("Shift"),
-		 nabi_server->statistics.shift);
+		 _("Total"), nabi_server->statistics.total,
+		 _("Space"), nabi_server->statistics.space,
+		 _("BackSpace"), nabi_server->statistics.backspace,
+		 _("Shift"), nabi_server->statistics.shift);
 
-	snprintf(choseong, sizeof(choseong),
-                 "%s\n"
-                 "\341\204\200: %-3d "
-                 "\341\204\201: %-3d "
-                 "\341\204\202: %-3d "
-                 "\341\204\203: %-3d "
-                 "\341\204\204: %-3d "
-                 "\341\204\205: %-3d "
-                 "\341\204\206: %-3d "
-                 "\341\204\207: %-3d "
-                 "\341\204\210: %-3d "
-                 "\341\204\211: %-3d\n"
-                 "\341\204\212: %-3d "
-                 "\341\204\213: %-3d "
-                 "\341\204\214: %-3d "
-                 "\341\204\215: %-3d "
-                 "\341\204\216: %-3d "
-                 "\341\204\217: %-3d "
-                 "\341\204\220: %-3d "
-                 "\341\204\221: %-3d "
-                 "\341\204\222: %-3d",
-		 _("Choseong"),                     
-                 nabi_server->statistics.jamo[0x00],
-                 nabi_server->statistics.jamo[0x01],
-                 nabi_server->statistics.jamo[0x02],
-                 nabi_server->statistics.jamo[0x03],
-                 nabi_server->statistics.jamo[0x04],
-                 nabi_server->statistics.jamo[0x05],
-                 nabi_server->statistics.jamo[0x06],
-                 nabi_server->statistics.jamo[0x07],
-                 nabi_server->statistics.jamo[0x08],
-                 nabi_server->statistics.jamo[0x09],
-                 nabi_server->statistics.jamo[0x0a],
-                 nabi_server->statistics.jamo[0x0b],
-                 nabi_server->statistics.jamo[0x0c],
-                 nabi_server->statistics.jamo[0x0d],
-                 nabi_server->statistics.jamo[0x0e],
-                 nabi_server->statistics.jamo[0x0f],
-                 nabi_server->statistics.jamo[0x10],
-                 nabi_server->statistics.jamo[0x11],
-                 nabi_server->statistics.jamo[0x12]);
-	
-	snprintf(jungseong, sizeof(jungseong),
-		 "%s\n"
-		 "\341\205\241: %-3d "
-		 "\341\205\242: %-3d "
-		 "\341\205\243: %-3d "
-		 "\341\205\244: %-3d "
-		 "\341\205\245: %-3d "
-		 "\341\205\246: %-3d "
-		 "\341\205\247: %-3d "
-		 "\341\205\250: %-3d "
-		 "\341\205\251: %-3d "
-		 "\341\205\252: %-3d\n"
-		 "\341\205\253: %-3d "
-		 "\341\205\254: %-3d "
-		 "\341\205\255: %-3d "
-		 "\341\205\256: %-3d "
-		 "\341\205\257: %-3d "
-		 "\341\205\260: %-3d "
-		 "\341\205\261: %-3d "
-		 "\341\205\262: %-3d "
-		 "\341\205\263: %-3d "
-		 "\341\205\264: %-3d "
-		 "\341\205\265: %-3d",
-		 _("Jungseong"),                    
-                 nabi_server->statistics.jamo[0x61],
-                 nabi_server->statistics.jamo[0x62],
-                 nabi_server->statistics.jamo[0x63],
-                 nabi_server->statistics.jamo[0x64],
-                 nabi_server->statistics.jamo[0x65],
-                 nabi_server->statistics.jamo[0x66],
-                 nabi_server->statistics.jamo[0x67],
-                 nabi_server->statistics.jamo[0x68],
-                 nabi_server->statistics.jamo[0x69],
-                 nabi_server->statistics.jamo[0x6a],
-                 nabi_server->statistics.jamo[0x6b],
-                 nabi_server->statistics.jamo[0x6c],
-                 nabi_server->statistics.jamo[0x6d],
-                 nabi_server->statistics.jamo[0x6e],
-                 nabi_server->statistics.jamo[0x6f],
-                 nabi_server->statistics.jamo[0x70],
-                 nabi_server->statistics.jamo[0x71],
-                 nabi_server->statistics.jamo[0x72],
-                 nabi_server->statistics.jamo[0x73],
-                 nabi_server->statistics.jamo[0x74],
-		 nabi_server->statistics.jamo[0x75]);
+	/* choseong */
+	g_string_append(str, _("Choseong"));
+	g_string_append_c(str, '\n');
+	for (i = 0x00; i <= 0x12; i++) {
+	    char label[8] = { '\0', };
+	    int len = g_unichar_to_utf8(0x1100 + i, label);
+	    label[len] = '\0';
+	    g_string_append_printf(str, "%s: %-3d ",
+				   label, nabi_server->statistics.jamo[i]);
+	    if ((i + 1) % 10 == 0)
+		g_string_append_c(str, '\n');
+	}
+	g_string_append(str, "\n\n");
 
-	snprintf(jongseong, sizeof(jongseong),
-		 "%s\n"
-                 "\341\206\250: %-3d "
-                 "\341\206\251: %-3d "
-                 "\341\206\252: %-3d "
-                 "\341\206\253: %-3d "
-                 "\341\206\254: %-3d "
-                 "\341\206\255: %-3d "
-                 "\341\206\256: %-3d "
-                 "\341\206\257: %-3d "
-                 "\341\206\260: %-3d "
-                 "\341\206\261: %-3d\n"
-                 "\341\206\262: %-3d "
-                 "\341\206\263: %-3d "
-                 "\341\206\264: %-3d "
-                 "\341\206\265: %-3d "
-                 "\341\206\266: %-3d "
-                 "\341\206\267: %-3d "
-                 "\341\206\270: %-3d "
-                 "\341\206\271: %-3d "
-                 "\341\206\272: %-3d "
-                 "\341\206\273: %-3d\n"
-                 "\341\206\274: %-3d "
-                 "\341\206\275: %-3d "
-                 "\341\206\276: %-3d "
-                 "\341\206\277: %-3d "
-                 "\341\207\200: %-3d "
-                 "\341\207\201: %-3d "
-                 "\341\207\202: %-3d",
-		 _("Jongseong"),
-		 nabi_server->statistics.jamo[0xa8],
-		 nabi_server->statistics.jamo[0xa9],
-		 nabi_server->statistics.jamo[0xaa],
-		 nabi_server->statistics.jamo[0xab],
-		 nabi_server->statistics.jamo[0xac],
-		 nabi_server->statistics.jamo[0xad],
-		 nabi_server->statistics.jamo[0xae],
-		 nabi_server->statistics.jamo[0xaf],
-		 nabi_server->statistics.jamo[0xb0],
-		 nabi_server->statistics.jamo[0xb1],
-		 nabi_server->statistics.jamo[0xb2],
-		 nabi_server->statistics.jamo[0xb3],
-		 nabi_server->statistics.jamo[0xb4],
-		 nabi_server->statistics.jamo[0xb5],
-		 nabi_server->statistics.jamo[0xb6],
-		 nabi_server->statistics.jamo[0xb7],
-		 nabi_server->statistics.jamo[0xb8],
-		 nabi_server->statistics.jamo[0xb9],
-		 nabi_server->statistics.jamo[0xba],
-		 nabi_server->statistics.jamo[0xbb],
-		 nabi_server->statistics.jamo[0xbc],
-                 nabi_server->statistics.jamo[0xbd],
-                 nabi_server->statistics.jamo[0xbe],
-                 nabi_server->statistics.jamo[0xbf],
-                 nabi_server->statistics.jamo[0xc0],
-                 nabi_server->statistics.jamo[0xc1],
-                 nabi_server->statistics.jamo[0xc2]);
+	/* jungseong */
+	g_string_append(str, _("Jungseong"));
+	g_string_append_c(str, '\n');
+	for (i = 0x61; i <= 0x75; i++) {
+	    char label[8] = { '\0', };
+	    int len = g_unichar_to_utf8(0x1100 + i, label);
+	    label[len] = '\0';
+	    g_string_append_printf(str, "%s: %-3d ",
+				   label, nabi_server->statistics.jamo[i]);
+	    if ((i - 0x60) % 10 == 0)
+		g_string_append_c(str, '\n');
+	}
+	g_string_append(str, "\n\n");
 
-	snprintf(buf, bufsize,
-		 "%s%s\n\n%s\n\n%s\n", general, choseong, jungseong, jongseong);
+	/* print only when a user have pressed any jonseong keys */
+	sum = 0;
+	for (i = 0xa8; i <= 0xc2; i++) {
+	    sum += nabi_server->statistics.jamo[i];
+	}
+
+	if (sum > 0) {
+	    /* jongseong */
+	    g_string_append(str, _("Jongseong"));
+	    g_string_append_c(str, '\n');
+	    for (i = 0xa8; i <= 0xc2; i++) {
+		char label[8] = { '\0', };
+		int len = g_unichar_to_utf8(0x1100 + i, label);
+		label[len] = '\0';
+		g_string_append_printf(str, "%s: %-3d ",
+				       label, nabi_server->statistics.jamo[i]);
+		if ((i - 0xa7) % 10 == 0)
+		    g_string_append_c(str, '\n');
+	    }
+	    g_string_append(str, "\n");
+	}
     }
 }
 
 static void
 on_about_statistics_clicked(GtkWidget *widget, gpointer parent)
 {
-    gchar stat_str[1536];
+    GString *str;
     GtkWidget *hbox;
     GtkWidget *stat_label = NULL;
     GtkWidget *dialog = NULL;
@@ -1018,14 +588,17 @@ on_about_statistics_clicked(GtkWidget *widget, gpointer parent)
     gtk_container_set_border_width(GTK_CONTAINER(hbox), 6);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox, TRUE, TRUE, 0);
 
-    get_statistic_string(stat_str, sizeof(stat_str));
-    stat_label = gtk_label_new(stat_str);
+    str = g_string_new(NULL);
+    get_statistic_string(str);
+    stat_label = gtk_label_new(str->str);
     gtk_label_set_selectable(GTK_LABEL(stat_label), TRUE);
     gtk_box_pack_start(GTK_BOX(hbox), stat_label, TRUE, TRUE, 6);
 
     gtk_widget_show_all(dialog);
     gtk_dialog_run(GTK_DIALOG(dialog));
     gtk_widget_destroy(dialog);
+
+    g_string_free(str, TRUE);
 }
 
 static void
@@ -1112,7 +685,7 @@ on_menu_about(GtkWidget *widget)
     gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
     gtk_table_attach_defaults(GTK_TABLE(server_info), label, 1, 2, 2, 3);
 
-    snprintf(buf, sizeof(buf), "%d", nabi_server->n_connected);
+    snprintf(buf, sizeof(buf), "%d", g_slist_length(nabi_server->connections));
     label = gtk_label_new(buf);
     gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
     gtk_table_attach_defaults(GTK_TABLE(server_info), label, 1, 2, 3, 4);
@@ -1172,8 +745,9 @@ on_menu_preference(GtkWidget *widget)
 static void
 on_menu_keyboard(GtkWidget *widget, gpointer data)
 {
-    nabi_app_set_keyboard((const char*)data);
+    nabi_app_set_hangul_keyboard((const char*)data);
     preference_window_update();
+    nabi_app_save_config();
 }
 
 static void
@@ -1189,8 +763,6 @@ create_menu(void)
     GtkWidget* menu_item;
     GtkWidget* image;
     GtkAccelGroup *accel_group;
-    NabiKeyboardTable *table;
-    GList *list;
     GSList *radio_group = NULL;
 
     accel_group = gtk_accel_group_new();
@@ -1230,30 +802,24 @@ create_menu(void)
 
     /* keyboard list */
     if (!nabi->status_only) {
-	list = nabi_server->keyboard_tables;
-	while (list != NULL) {
-	    table = (NabiKeyboardTable*)list->data;
+	int i = 0;
+	const NabiHangulKeyboard* keyboards;
+	keyboards = nabi_server_get_hangul_keyboard_list(nabi_server);
+	while (keyboards[i].id != NULL) {
+	    const char* id = keyboards[i].id;
+	    const char* name = keyboards[i].name;
 	    menu_item = gtk_radio_menu_item_new_with_label(radio_group,
-							   table->name);
+							   _(name));
 	    radio_group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(menu_item));
 	    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
 	    gtk_widget_show(menu_item);
 	    g_signal_connect(G_OBJECT(menu_item), "activate",
-			     G_CALLBACK(on_menu_keyboard), table->name);
-	    if (strcmp(table->name, nabi->keyboard_table_name) == 0)
+			     G_CALLBACK(on_menu_keyboard), (gpointer)id);
+	    if (strcmp(id, nabi->config->hangul_keyboard) == 0)
 		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_item),
 					       TRUE);
-	    list = list->next;
+	    i++;
 	}
-	/* do not add dvorak option to menu
-	menu_item = gtk_check_menu_item_new_with_label(_("Dvorak layout"));
-	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu_item),
-				       nabi->dvorak);
-	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
-	gtk_widget_show(menu_item);
-	g_signal_connect_swapped(G_OBJECT(menu_item), "activate",
-				 G_CALLBACK(on_menu_dvorak), menu_item);
-				 */
 
 	/* separator */
 	menu_item = gtk_separator_menu_item_new();
@@ -1477,44 +1043,56 @@ on_main_window_realized(GtkWidget *widget, gpointer data)
 	nabi_server_set_mode_info_cb(nabi_server, nabi_set_input_mode_info);
 }
 
+static GdkPixbuf*
+load_icon(const char* theme, const char* name, const char** default_xpm)
+{
+    char* path;
+    char* filename;
+    GError *error = NULL;
+    GdkPixbuf *pixbuf;
+
+    filename = g_strconcat(name, ".svg", NULL);
+    path = g_build_filename(NABI_THEMES_DIR, theme, filename, NULL);
+    pixbuf = gdk_pixbuf_new_from_file(path, &error);
+    if (pixbuf != NULL)
+	goto done;
+
+    g_free(filename);
+    g_free(path);
+    if (error != NULL) {
+	g_error_free(error);
+	error = NULL;
+    }
+
+    filename = g_strconcat(name, ".png", NULL);
+    path = g_build_filename(NABI_THEMES_DIR, theme, filename, NULL);
+    pixbuf = gdk_pixbuf_new_from_file(path, &error);
+    if (pixbuf == NULL) {
+	if (error != NULL) {
+	    g_print("Error on reading image file: %s\n", error->message);
+	    g_error_free(error);
+	}
+	pixbuf = gdk_pixbuf_new_from_xpm_data(none_default_xpm);
+    }
+
+done:
+    g_free(filename);
+    g_free(path);
+    if (error != NULL)
+	g_error_free(error);
+
+    return pixbuf;
+}
+
 static void
 load_base_icons(const gchar *theme)
 {
-    gchar *path;
-    GError *gerror = NULL;
-
     if (theme == NULL)
-    	theme = "SimplyRed";
+    	theme = "Jini";
 
-    path = g_build_filename(NABI_THEMES_DIR, theme, "none.png", NULL);
-    none_pixbuf = gdk_pixbuf_new_from_file(path, &gerror);
-    g_free(path);
-    if (gerror != NULL) {
-	g_print("Error on reading image file: %s\n", gerror->message);
-	g_error_free(gerror);
-	gerror = NULL;
-	none_pixbuf = gdk_pixbuf_new_from_xpm_data(none_default_xpm);
-    }
-
-    path = g_build_filename(NABI_THEMES_DIR, theme, "hangul.png", NULL);
-    hangul_pixbuf = gdk_pixbuf_new_from_file(path, &gerror);
-    g_free(path);
-    if (gerror != NULL) {
-	g_print("Error on reading image file: %s\n", gerror->message);
-	g_error_free(gerror);
-	gerror = NULL;
-	hangul_pixbuf = gdk_pixbuf_new_from_xpm_data(hangul_default_xpm);
-    }
-
-    path = g_build_filename(NABI_THEMES_DIR, theme, "english.png", NULL);
-    english_pixbuf = gdk_pixbuf_new_from_file(path, &gerror);
-    g_free(path);
-    if (gerror != NULL) {
-	g_print("Error on reading image file: %s\n", gerror->message);
-	g_error_free(gerror);
-	gerror = NULL;
-	english_pixbuf = gdk_pixbuf_new_from_xpm_data(english_default_xpm);
-    }
+    none_pixbuf = load_icon(theme, "none", none_default_xpm);
+    english_pixbuf = load_icon(theme, "english", english_default_xpm);
+    hangul_pixbuf = load_icon(theme, "hangul", hangul_default_xpm);
 }
 
 static gboolean
@@ -1541,7 +1119,7 @@ create_tray_icon(gpointer data)
 			 _("Hangul input method: Nabi"
 			   " - You can input hangul using this program"));
 
-    load_base_icons(nabi->theme);
+    load_base_icons(nabi->config->theme);
     create_resized_icons(nabi->icon_size);
 
     hbox = gtk_hbox_new(TRUE, 0);
@@ -1571,7 +1149,6 @@ nabi_app_create_main_widget(void)
     GtkWidget *window;
     GtkWidget *frame;
     GtkWidget *ebox;
-    GList *list = NULL;
 
     window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
     gtk_window_set_title(GTK_WINDOW(window), _("Nabi: None"));
@@ -1579,7 +1156,7 @@ nabi_app_create_main_widget(void)
     gtk_window_set_skip_pager_hint(GTK_WINDOW(window), TRUE);
     gtk_window_set_skip_taskbar_hint(GTK_WINDOW(window), TRUE);
     gtk_window_stick(GTK_WINDOW(window));
-    gtk_window_move(GTK_WINDOW(window), nabi->x, nabi->y);
+    gtk_window_move(GTK_WINDOW(window), nabi->config->x, nabi->config->y);
     g_signal_connect_after(G_OBJECT(window), "realize",
 	    		   G_CALLBACK(on_main_window_realized), NULL);
     g_signal_connect(G_OBJECT(window), "destroy",
@@ -1605,9 +1182,11 @@ nabi_app_create_main_widget(void)
 
     g_idle_add(create_tray_icon, NULL);
 
-    list = g_list_prepend(list, default_icon);
-    gtk_window_set_default_icon_list(list);
-    g_list_free(list);
+    if (default_icon != NULL) {
+	GList *list = g_list_prepend(NULL, default_icon);
+	gtk_window_set_default_icon_list(list);
+	g_list_free(list);
+    }
 
     return window;
 }
@@ -1618,57 +1197,26 @@ nabi_app_set_theme(const gchar *name)
     load_base_icons(name);
     create_resized_icons(nabi->icon_size);
 
-    g_free(nabi->theme);
-    nabi->theme = g_strdup(name);
-    nabi_save_config_file();
+    g_free(nabi->config->theme);
+    nabi->config->theme = g_strdup(name);
 }
 
 void
-nabi_app_set_dvorak(gboolean state)
+nabi_app_set_hangul_keyboard(const char *id)
 {
-    nabi->dvorak = state;
-    nabi_server_set_dvorak(nabi_server, nabi->dvorak);
-    nabi_save_config_file();
+    nabi_server_set_hangul_keyboard(nabi_server, id);
+    g_free(nabi->config->hangul_keyboard);
+    if (id == NULL)
+	nabi->config->hangul_keyboard = NULL;
+    else
+	nabi->config->hangul_keyboard = g_strdup(id);
 }
 
 void
-nabi_app_set_keyboard(const char *name)
+nabi_app_save_config()
 {
-    if (name != NULL) {
-	nabi_server_set_keyboard_table(nabi_server, name);
-	g_free(nabi->keyboard_table_name);
-	nabi->keyboard_table_name = g_strdup(name);
-	nabi_save_config_file();
-    }
-}
-
-void
-nabi_app_set_candidate_font(const char *font)
-{
-    if (font != NULL) {
-	g_free(nabi->candidate_font);
-	nabi->candidate_font = g_strdup(font);
-	nabi_server_set_candidate_font(nabi_server, nabi->candidate_font);
-	nabi_save_config_file();
-    }
-}
-
-void
-nabi_app_set_trigger_keys(char **keys)
-{
-    g_free(nabi->trigger_keys);
-    nabi->trigger_keys = g_strjoinv(",", keys);
-    nabi_server_set_trigger_keys(nabi_server, keys);
-    nabi_save_config_file();
-}
-
-void
-nabi_app_set_candidate_keys(char **keys)
-{
-    g_free(nabi->candidate_keys);
-    nabi->candidate_keys = g_strjoinv(",", keys);
-    nabi_server_set_candidate_keys(nabi_server, keys);
-    nabi_save_config_file();
+    if (nabi != NULL && nabi->config != NULL)
+	nabi_config_save(nabi->config);
 }
 
 /* vim: set ts=8 sts=4 sw=4 : */
